import atexit
import os
import re
import signal
import sys
import threading
import time
from typing import Any, Dict, List, Optional, Union

from urllib3.exceptions import MaxRetryError

from openapi_client.models.experiment_metric_entry import ExperimentMetricEntry
from openapi_client.models.experiment_metrics_update_api_input import (
    ExperimentMetricsUpdateAPIInput,
)
from openapi_client.models.local_experiment_finish_api_input import (
    LocalExperimentFinishAPIInput,
)
from openapi_client.models.response_experiment_info import ResponseExperimentInfo
from vessl.cli._util import Endpoint
from vessl.integration.tensorboard import integrate_tensorboard
from vessl.internal.collector import (
    Collector,
    IOCollector,
    SystemMetricCollector,
    UserMetricCollector,
)
from vessl.internal.progress_updater import ProgressUpdater
from vessl.internal.vessl_hyperparameters import (
    Hyperparameters,
    HyperparametersController,
)
from vessl.util import logger
from vessl.util.api import VesslApi
from vessl.util.audio import Audio
from vessl.util.constant import (
    VESSL_MEDIA_PATH,
    VESSL_PLOTS_FILETYPE_AUDIO,
    VESSL_PLOTS_FILETYPE_IMAGE,
)
from vessl.util.exception import (
    InvalidExperimentError,
    InvalidTypeError,
    VesslApiException,
)
from vessl.util.image import Image

MODE_TEST = "TEST"
MODE_NOT_STARTED = "NOT_STARTED"
MODE_LOCAL = "LOCAL"
MODE_MANAGED = "MANAGED"

SEND_INTERVAL_IN_SEC = 10

METRIC_KEY_REGEX = re.compile("^[a-zA-Z0-9/_-]+$")


class Sender(object):
    def __init__(self, api: VesslApi, experiment_id: int, collectors: List[Collector]):
        self._api = api
        self._experiment_id: int = experiment_id
        self._thread = threading.Thread(target=self._thread_body, daemon=True)
        self._exit = threading.Event()
        self._collectors = collectors

    def stop(self):
        for c in self._collectors:
            c.stop()

        self._exit.set()
        self._thread.join()

    def start(self):
        for c in self._collectors:
            c.start()
        self._thread.start()

    def _thread_body(self):
        while not self._exit.is_set():
            self._send()
            self._exit.wait(timeout=SEND_INTERVAL_IN_SEC)
        self._send()

    def _send(self):
        pairs = [(c, c.collect()) for c in self._collectors]
        for c, m in pairs:
            logger.debug(f"{c} / {m}", str(c), len(m))
        payload = [m for _, metrics in pairs for m in metrics]
        logger.debug(f"Sending {len(payload)} payloads")

        try:
            res = self._api.experiment_metrics_update_api(
                self._experiment_id,
                experiment_metrics_update_api_input=ExperimentMetricsUpdateAPIInput(
                    metrics=payload
                ),
            )
            if res.rejected:
                logger.warning(f"{res.rejected} payloads(s) were rejected.")

            for c, m in pairs:
                c.truncate(len(m))

        except (MaxRetryError, VesslApiException) as e:
            logger.exception("Failed to send metrics to server", exc_info=e)
        except Exception as e:
            logger.exception("Unexpected error", exc_info=e)


class VesslRun(object):
    class ExitHook(object):
        def __init__(self, orig_exit):
            self.orig_exit = orig_exit
            self.exit_code = 0

        def exit(self, code=0):
            self.exit_code = code
            self.orig_exit(code)

    __slots__ = [
        "api",
        "_mode",
        "_collectors",
        "_sender",
        "_progress_updater",
        "hyperparameters",
        "_experiment",
        "_logger",
        "_user_metric_collector",
        "_exit_hook",
        "_tensorboard_collector",
    ]

    def __init__(self) -> None:
        self.api = VesslApi()
        self._experiment = self._get_experiment_from_environment()
        self.hyperparameters = HyperparametersController(self.api, self._experiment)
        self._mode = MODE_NOT_STARTED if self._experiment is None else MODE_MANAGED

        self._user_metric_collector = UserMetricCollector()
        self._exit_hook = self.ExitHook(sys.exit)

    def _get_experiment_from_environment(self) -> Optional[ResponseExperimentInfo]:
        """Detect experiment from environment variables

        In a Vessl-managed experiment, these variables will be defined.
        """
        experiment_id = os.environ.get("VESSL_EXPERIMENT_ID", None)
        access_token = os.environ.get("VESSL_ACCESS_TOKEN", None)

        if experiment_id is None or access_token is None:
            return None

        self.api.configure_access_token(access_token)
        try:
            return self.api.experiment_read_by_idapi(experiment_id=experiment_id)
        except VesslApiException:
            return None

    def _get_experiment_from_args(
        self,
        experiment_number,
        message: str = None,
        hyperparameters: Hyperparameters = None,
    ) -> ResponseExperimentInfo:
        """Get or create a local experiment

        If experiment is specified, use it. Otherwise, create a new experiment.
        """
        # Create a new experiment
        if experiment_number is None:
            from vessl.experiment import create_local_experiment

            experiment = create_local_experiment(message=message, hyperparameters=hyperparameters)
            logger.debug(f"Created experiment {experiment.id}")
            return experiment

        # Continue with previous experiment
        from vessl.experiment import read_experiment

        experiment = read_experiment(experiment_number)
        if not experiment.is_local or experiment.local_execution_spec is None:
            raise InvalidExperimentError(
                f"{experiment.number}: cannot use Vessl-managed experiment."
            )
        if experiment.status != "running":
            raise InvalidExperimentError(f"{experiment.number}: experiment must be running.")

        if message is not None and experiment.message != message:
            from vessl.experiment import update_experiment

            experiment = update_experiment(experiment.number, message=message)
            logger.debug(f'Updated experiment message to: "{message}"')

        return experiment

    def _signal_handler(self, signo, frames):
        sys.exit(130)  # job was terminated by the owner

    def _start(self):
        """Start sender and register hooks"""
        self._sender.start()

        sys.exit = self._exit_hook.exit
        atexit.register(self._stop)
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)

    def _stop(self):
        """Stop sender and restore hooks"""
        if self._mode != MODE_LOCAL:
            return

        self._sender.stop()

        sys.exit = self._exit_hook.orig_exit
        self.api.local_experiment_finish_api(
            self._experiment.id,
            local_experiment_finish_api_input=LocalExperimentFinishAPIInput(
                exit_code=self._exit_hook.exit_code
            ),
        )

    def _send_without_collector(self, payloads):
        """Send metrics without using the collector

        In a Vessl-managed experiment, metrics are sent immediately instead of being queued.
        """
        assert self._mode == MODE_MANAGED
        res = self.api.experiment_metrics_update_api(
            self._experiment.id,
            experiment_metrics_update_api_input=ExperimentMetricsUpdateAPIInput(metrics=payloads),
        )
        if res.rejected:
            logger.warning(f"{res.rejected} payloads(s) were rejected.")

    def init(
        self,
        experiment_number=None,
        message: str = None,
        tensorboard: bool = False,
        hp: dict = None,
        **kwargs,
    ):
        """Main function to setup Vessl in a local setting

        If this is a Vessl-managed experiment or vessl.init has already been called,
        this will do nothing.

        Args:
            experiment_number (str | int): experiment name or number
            message (str): experiment message
            tensorboard (bool): enable tensorboard integration. It is important to note
              that `vessl.init` must be called **before** initializing the writer
              (`tf.summary.create_file_writer` for TF2, SummaryWriter for PyTorch, etc).
            hp (dict): hyperparameters
        """
        if self._mode == MODE_NOT_STARTED:
            print(f"Initializing a new experiment...")
            self.api.set_configurations(
                organization_name=kwargs.get("organization_name"),
                project_name=kwargs.get("project_name"),
            )

            if hp:
                self.hyperparameters.update_items(hp)

            self._experiment = self._get_experiment_from_args(
                experiment_number,
                message,
                self.hyperparameters.as_list_of_dict(),
            )
            self._mode = MODE_LOCAL
            self.hyperparameters.configure(self.api, self._experiment)

            print(
                f"Connected to experiment #{self._experiment.number}.\n"
                f"For more info: {Endpoint.experiment.format(self._experiment.organization.name, self._experiment.project.name, self._experiment.number)}"
            )

            # For testing
            if "is_test" in kwargs and kwargs["is_test"]:
                return

            gpu_count = self._experiment.local_execution_spec.gpu_count or 0
            self._user_metric_collector = UserMetricCollector()
            collectors = [
                IOCollector(),
                SystemMetricCollector(gpu_count),
                self._user_metric_collector,
            ]
            self._sender = Sender(self.api, self._experiment.id, collectors)
            self._start()

        elif self._mode == MODE_LOCAL:
            print(
                f"Already connected to experiment #{self._experiment.number}.\n"
                "To start a new experiment, call `vessl.finish()` first and `vessl.init()` again."
            )

        if self._mode == MODE_MANAGED:
            if hp is not None:
                self.hyperparameters.update(hp)

            if message is not None and self._experiment.message != message:
                from vessl.experiment import update_experiment

                self._experiment = update_experiment(
                    self._experiment.number,
                    message=message,
                    organization_name=self._experiment.organization.name,
                    project_name=self._experiment.project.name,
                )
                logger.debug(f'Updated experiment.message to: "{message}"')
            else:
                logger.debug(
                    f'Connected to experiment #{self._experiment.number}. Message: "{self._experiment.message}"'
                )

        if tensorboard:
            integrate_tensorboard()

    def upload(self, path: str):
        """Upload output files

        Args:
            path (str): path to upload
        """
        if self._mode == MODE_NOT_STARTED:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        from vessl.experiment import upload_experiment_output_files

        upload_experiment_output_files(self._experiment.number, path)

    def finish(self):
        """Teardown Vessl settings

        Use this function to stop tracking your experiment mid-script. If not called,
        tracking is stopped automatically upon exit.

        Args:
            path (str): path to upload
        """
        if self._mode == MODE_NOT_STARTED:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        if self._mode == MODE_MANAGED:
            return

        self._stop()
        experiment_number = self._experiment.number
        self._mode = MODE_NOT_STARTED
        self._experiment = None
        print(f"Experiment {experiment_number} completed.")

    def log(
        self,
        payload: Dict[str, Any],
        step: Optional[int] = None,
        ts: Optional[float] = None,
    ):
        """Log metrics to Vessl

        Args:
            payload (Dict[str, Any]): to log a scalar, value should be a number. To
                log an image, pass a single image or a list of images (type `vessl.util.image.Image`).
            step (int): step.
        """
        if self._mode == MODE_NOT_STARTED:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        if ts is None:
            ts = time.time()

        scalar_dict = {}
        media_dict = {}

        for k, v in payload.items():
            if isinstance(v, list):
                if all(isinstance(i, Image) for i in v) or all(isinstance(i, Audio) for i in v):
                    media_dict[k] = v
            elif isinstance(v, Image) or isinstance(v, Audio):
                media_dict[k] = [v]
            elif isinstance(v, int) or isinstance(v, float):
                scalar_dict[k] = v
            else:
                try:
                    scalar_dict[k] = float(v)
                except ValueError:
                    logger.error(
                        f"Could not convert '{type(v).__name__}' {v} to float. "
                        f"(allowed types: Image, Audio, int, float)"
                    )

        # Update step if step is specified. If a scalar is defined but step wasn't,
        # step will be autoincremented.
        if scalar_dict or step is not None:
            self._user_metric_collector.handle_step(step)

        if media_dict:
            self._update_media(media_dict, ts)

        if scalar_dict:
            self._update_metrics(scalar_dict, ts)

    # This should mirror app/influx/metric_schema.go > `isValidMetricKey`
    def _is_metric_key_valid(self, key: str):
        if not METRIC_KEY_REGEX.match(key):
            return False
        if key.startswith("/") or key.endswith("/"):
            return False
        for i in range(1, len(key)):
            if key[i] == "/" and key[i - 1] == "/":
                return False
        return True

    def _update_metrics(self, payload: Dict[str, Any], ts: float):
        invalid_keys = [k for k in payload.keys() if not self._is_metric_key_valid(k)]
        if invalid_keys:
            logger.warning(
                f"Invalid metric keys: {' '.join(invalid_keys)}. This payload will be rejected."
            )

        payloads = [self._user_metric_collector.build_metric_payload(payload, ts)]
        if self._mode == MODE_MANAGED:
            self._send_without_collector(payloads)
            return self._user_metric_collector.step
        return self._user_metric_collector.log_metrics(payloads)

    def _update_media(
        self,
        payload: Dict[str, Union[List[Image], List[Audio]]],
        ts: float,
    ):
        media_type, set_media_type = None, False
        path_to_caption = {}
        for media in payload.values():
            for medium in media:
                path = os.path.basename(medium.path)
                path_to_caption[path] = medium.caption
                if not set_media_type:
                    if not isinstance(medium, Image) and not isinstance(medium, Audio):
                        raise InvalidTypeError(f"Invalid payload type error: {type(medium)}")
                    else:
                        media_type = type(medium).__name__
                        set_media_type = True

        from vessl.volume import copy_volume_file

        files = copy_volume_file(
            source_volume_id=None,
            source_path=os.path.join(VESSL_MEDIA_PATH, "."),
            dest_volume_id=self._experiment.experiment_plot_volume,
            dest_path="/",
        )

        for media in payload.values():
            for medium in media:
                medium.flush()

        plot_files = []
        if files:
            plot_files = [
                {
                    "step": None,
                    "path": file.get("path"),
                    "caption": path_to_caption[file.get("path")],
                    "timestamp": ts,
                }
                for file in files
                if file.get("path") in path_to_caption
            ]

        payloads: List[ExperimentMetricEntry] = []
        for f in plot_files:
            payload = {}
            if media_type == Image.__name__:
                payload = {VESSL_PLOTS_FILETYPE_IMAGE: f}
            elif media_type == Audio.__name__:
                payload = {VESSL_PLOTS_FILETYPE_AUDIO: f}
            payloads.append(self._user_metric_collector.build_media_payload(payload, ts))

        if self._mode == MODE_MANAGED:
            self._send_without_collector(payloads)
        else:
            self._user_metric_collector.log_media(payloads)

    def progress(self, value: float):
        """Update experiment progress

        Args:
            value (float): progress value as a decimal between 0 and 1
        """
        if self._experiment is None:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        if not 0 < value <= 1:
            logger.warning(f"Invalid progress value {value}. (0 < value <= 1)")
            return

        if not hasattr(self, "_progress_updater"):
            # Do not initialize in init() since ProgressUpdater might not be used at all
            self._progress_updater = ProgressUpdater(self.api, self._experiment.id)
            self._progress_updater.start()

        self._progress_updater.update(value)
        logger.debug(f"Experiment progress: {value}")



import atexit
import importlib
import io
import os
import threading
from typing import Optional

import vessl
from vessl.util import logger
from vessl.util.audio import Audio
from vessl.util.common import get_module
from vessl.util.image import Image

MODE_TENSORFLOW2 = "mode-tf2"
MODE_OTHERS = "mode-others"

# Tensorboard event types
EVENT_TYPE_COMPRESSED_HISTOGRAMS = "distributions"
EVENT_TYPE_HISTOGRAMS = "histograms"
EVENT_TYPE_IMAGES = "images"
EVENT_TYPE_AUDIO = "audio"
EVENT_TYPE_SCALARS = "scalars"
EVENT_TYPE_TENSORS = "tensors"

# EventAccumulator argument (0 = keep all, N = keep N)
SIZE_GUIDANCE = {
    EVENT_TYPE_COMPRESSED_HISTOGRAMS: 1,
    EVENT_TYPE_IMAGES: 0,
    EVENT_TYPE_AUDIO: 0,
    EVENT_TYPE_SCALARS: 0,
    EVENT_TYPE_HISTOGRAMS: 1,
    EVENT_TYPE_TENSORS: 1,
}

READ_INTERVAL_IN_SEC = 1


class TensorboardCollector:
    """TensorboardCollector checks for new tensorboard events written to logdir and
    logs them to experiment metrics.

    There are two modes: `MODE_TENSORFLOW2` (TF2) and `MODE_OTHERS` (TF1, PyTorch, etc).
    `MODE_OTHERS` uses tensorboard's `EventAccumulator` to read from the event file.
    `MODE_TENSORFLOW2` uses tensorflow's `summary_iterator` because TF2 records events
      in a different way.
    """

    __slots__ = [
        "_exit",
        "_thread",
        "_tf_dict",
        "_event_accumulator",
    ]

    def start(self):
        self._start_thread()

    @property
    def _mode(self):
        if hasattr(self, "_tf_dict"):
            return MODE_TENSORFLOW2
        if hasattr(self, "_event_accumulator"):
            return MODE_OTHERS
        return ""

    def _is_initialized(self):
        return self._mode in (MODE_TENSORFLOW2, MODE_OTHERS)

    def _start_thread(self):
        def exit_fn(*args):
            self._exit.set()
            self._thread.join()
            if not self._is_initialized():
                logger.warn("No tensorboard writer was detected during the run.")

        atexit.register(exit_fn)

        self._exit = threading.Event()
        self._thread = threading.Thread(target=self._thread_body, daemon=True)
        self._thread.start()

    def _thread_body(self):
        while not self._exit.is_set():
            self._main()
            self._exit.wait(timeout=READ_INTERVAL_IN_SEC)
        self._main()

    def _main(self):
        if not self._is_initialized():
            # Wait for initialization
            return

        if self._mode == MODE_TENSORFLOW2:
            self._main_tensorflow2()

        elif self._mode == MODE_OTHERS:
            self._main_others()

    def _main_tensorflow2(self):
        """Main thread body for MODE_TENSORFLOW2"""
        # Skip already seen events
        for _ in range(self._tf_dict["start_index"]):
            try:
                next(self._tf_dict["summary_iterator"])
            except StopIteration:
                continue

        events = []
        for e in self._tf_dict["summary_iterator"]:
            events.append(e)

        self._tf_dict["start_index"] += len(events)
        for e in events:
            self._handle_event(e)

    def _handle_event(self, event):
        for val_struct in event.summary.value:
            event_type = val_struct.metadata.plugin_data.plugin_name

            if event_type == EVENT_TYPE_SCALARS:
                event_value = self._tf_dict["make_ndarray_fn"](val_struct.tensor).item()

            elif event_type == EVENT_TYPE_IMAGES:
                PILImage = get_module(
                    "PIL.Image",
                    required='Pillow package is required. Run "pip install Pillow".',
                )
                # First two tensors are height and width, image bytestrings start
                # from third tensor.
                event_value = [
                    Image(data=PILImage.open(io.BytesIO(bs)), caption=val_struct.tag)
                    for bs in val_struct.tensor.string_val[2:]
                ]

            elif event_type == EVENT_TYPE_AUDIO:
                wave = get_module(
                    "wave", required='wave package is required. Run "pip install wave".'
                )
                np = get_module(
                    "numpy",
                    required='numpy package is required. Run "pip install numpy".',
                )

                event_value = []
                for bs in val_struct.tensor.string_val:
                    if not bs:
                        continue
                    wave_object = wave.open(io.BytesIO(bs))
                    data = np.frombuffer(
                        wave_object.readframes(wave_object.getnframes()),
                        dtype="float32",
                    )
                    vessl_audio = Audio(
                        data,
                        sample_rate=wave_object.getframerate(),
                        caption=val_struct.tag,
                    )
                    event_value.append(vessl_audio)

            else:
                continue

            payload = {val_struct.tag: event_value}
            vessl.log(payload=payload, step=event.step, ts=event.wall_time)

    def _main_others(self):
        """Main thread body for MODE_OTHERS"""
        try:
            self._event_accumulator.Reload()  # Loads each event at most once
        except Exception:
            # No metrics have been saved yet
            return

        event_tags = self._event_accumulator.Tags()
        # ex. {"scalars": ["loss", "accuracy"], "images": ["caption"]}

        self._handle_scalars(event_tags[EVENT_TYPE_SCALARS])
        self._handle_images(event_tags[EVENT_TYPE_IMAGES])
        self._handle_audios(event_tags[EVENT_TYPE_AUDIO])

    def _handle_scalars(self, tags):
        for tag in tags:
            scalars = self._event_accumulator.Scalars(tag)
            self._flush_scalars(tag)  # Flush right away
            for scalar in scalars:
                self._log_scalar(tag, scalar)

    def _flush_scalars(self, tag):
        with self._event_accumulator.scalars._mutex:
            scalars = self._event_accumulator.scalars
        with scalars._buckets[tag]._mutex:
            scalars._buckets[tag].items = []

    def _log_scalar(self, tag, scalar):
        payload = {tag: scalar.value}
        vessl.log(payload=payload, step=scalar.step, ts=scalar.wall_time)

    def _handle_images(self, tags):
        for tag in tags:
            images = self._event_accumulator.Images(tag)
            self._flush_images(tag)  # Flush right away
            for image in images:
                self._log_image(tag, image)

    def _flush_images(self, tag):
        with self._event_accumulator.images._mutex:
            images = self._event_accumulator.images
        with images._buckets[tag]._mutex:
            images._buckets[tag].items = []

    def _log_image(self, tag, image):
        PILImage = get_module(
            "PIL.Image",
            required='Pillow package is required. Run "pip install Pillow".',
        )

        vessl_image = Image(data=PILImage.open(io.BytesIO(image.encoded_image_string)), caption=tag)

        payload = {tag: vessl_image}
        vessl.log(payload=payload, step=image.step, ts=image.wall_time)

    def _handle_audios(self, tags):
        for tag in tags:
            audios = self._event_accumulator.Audio(tag)
            self._flush_audios(tag)  # Flush right away
            for audio in audios:
                self._log_audio(tag, audio)

    def _flush_audios(self, tag):
        with self._event_accumulator.audios._mutex:
            audios = self._event_accumulator.audios
        with audios._buckets[tag]._mutex:
            audios._buckets[tag].items = []

    def _log_audio(self, tag, audio):
        wave = get_module("wave", required='wave package is required. Run "pip install wave".')
        np = get_module("numpy", required='numpy package is required. Run "pip install numpy".')

        wave_object = wave.open(io.BytesIO(audio.encoded_audio_string))
        data = np.frombuffer(wave_object.readframes(wave_object.getnframes()), dtype="float32")
        vessl_audio = Audio(data, sample_rate=int(audio.sample_rate), caption=tag)

        payload = {tag: vessl_audio}
        vessl.log(payload=payload, step=audio.step, ts=audio.wall_time)

    def set_logdir(self, logdir, mode):
        """Called when tensorboard writer is detected. Only the first logdir will be used."""
        if not self._is_initialized():
            logger.info(f"Tensorboard logdir detected: {logdir}.")
            if mode == MODE_TENSORFLOW2:
                self._initialize_mode_tensorflow2(logdir)
            elif mode == MODE_OTHERS:
                self._initialize_mode_others(logdir)
        else:
            logger.debug(f"Cannot use multiple tensorboard logdirs. {logdir} will be ignored.")

    def _initialize_mode_tensorflow2(self, logdir):
        try:
            from tensorflow import make_ndarray
            from tensorflow.compat.v1.train import summary_iterator
        except Exception:
            logger.error(f"Could not import tensorflow. Failed to integrate tensorboard.")
            return

        # Get most recent event file. The event file for this run will have been just
        # created (in `tf.python.ops.gen_summary_ops.create_summary_file_writer`).
        path = mtime = None
        for dirpath, _, file_names in os.walk(logdir):
            for file_name in file_names:
                file_path = os.path.join(dirpath, file_name)
                file_mtime = os.path.getmtime(file_path)
                if "tfevents" in file_name and (mtime is None or mtime < file_mtime):
                    path, mtime = file_path, file_mtime

        if path is None:
            logger.info(f"No event file found in logdir: {logdir}.")
            return

        # Dict for MODE_TENSORFLOW2 objects
        self._tf_dict = {
            "make_ndarray_fn": make_ndarray,
            "summary_iterator": summary_iterator(path),
            "start_index": 0,
        }

    def _initialize_mode_others(self, logdir):
        try:
            from tensorboard.backend.event_processing.event_accumulator import (
                EventAccumulator,
            )
        except Exception:
            logger.error(
                "Could not import tensorboard. (Please install using `pip install tensorboard` first.) Failed to integrate tensorboard.",
            )
        self._event_accumulator = EventAccumulator(logdir, size_guidance=SIZE_GUIDANCE)

        try:
            self._event_accumulator.Reload()
        except Exception:
            # Logdir doesn't exist yet
            return

        # Disregard preexisting tensorboard logs
        for type, tags in self._event_accumulator.Tags().items():
            if type == EVENT_TYPE_SCALARS:
                for tag in tags:
                    self._flush_scalars(tag)
            if type == EVENT_TYPE_IMAGES:
                for tag in tags:
                    self._flush_images(tag)
            if type == EVENT_TYPE_AUDIO:
                for tag in tags:
                    self._flush_audios(tag)

    def close(self):
        """Joins thread to stop collecting. Used by tests."""
        self._exit.set()
        self._thread.join()
        return


# Global variable to ensure only one `TensorboardCollector` is active
tensorboard_collector_instance: Optional[TensorboardCollector] = None


def integrate_tensorboard():
    """Integrate tensorboard

    This method is called from `vessl.init` and can also be called manually."""
    global tensorboard_collector_instance

    if tensorboard_collector_instance is not None:
        logger.warning("Tensorboard is already enabled.")
        return

    tensorboard_collector_instance = TensorboardCollector()
    tensorboard_collector_instance.start()

    _patch_create_summary_file_writer(tensorboard_collector_instance)
    _patch_event_file_writers(tensorboard_collector_instance)


def _patch_create_summary_file_writer(tc):
    module_name = "tensorflow.python.ops.gen_summary_ops"
    try:
        module = importlib.import_module(module_name)
    except Exception as e:
        logger.debug(f"Module {module_name} not found. Skipping patch...")
        return

    _create_summary_file_writer = module.create_summary_file_writer

    def custom_create_summary_file_writer(*args, **kwargs):
        ret = _create_summary_file_writer(*args, **kwargs)
        logdir = (
            kwargs["logdir"].numpy().decode("utf8")
            if hasattr(kwargs["logdir"], "numpy")
            else kwargs["logdir"]
        )
        # Calling `set_logdir` after `_create_summary_file_writer` ensures that the
        # event file has already been created. This is important because we use the
        # most recent file in `logdir` as our event file.
        tc.set_logdir(logdir, MODE_TENSORFLOW2)
        return ret

    module.create_summary_file_writer = custom_create_summary_file_writer
    logger.debug(f"Module {module_name} was successfully patched.")


def _patch_event_file_writers(tc):
    module_names = [
        "tensorflow.python.summary.writer.writer",  # TF1
        "tensorboard.summary.writer.event_file_writer",
        "torch.utils.tensorboard.writer",
        "tensorboardX.writer",
    ]

    for module_name in module_names:
        try:
            module = importlib.import_module(module_name)
        except Exception as e:
            logger.debug(f"Module {module_name} not found. Skipping patch...")
            continue

        _patch_event_file_writer(module, tc)
        logger.debug(f"Module {module_name} was successfully patched.")


def _patch_event_file_writer(module, tc):
    _EventFileWriter = module.EventFileWriter

    class CustomEventFileWriter(_EventFileWriter):
        def __init__(self, *args, **kwargs):
            # self.log_dir is set in super().__init__
            super(CustomEventFileWriter, self).__init__(*args, **kwargs)

            logdir = self.get_logdir()
            tc.set_logdir(logdir, MODE_OTHERS)

    module.EventFileWriter = CustomEventFileWriter



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmKernelCluster(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'default_storage_id': 'int',
        'edges': 'OrmKernelClusterEdges',
        'id': 'int',
        'immutable_slug': 'str',
        'influx_bucket_name': 'str',
        'is_custom_resource_spec_allowed': 'bool',
        'is_vessl_managed': 'bool',
        'kubernetes_container_runtime': 'str',
        'kubernetes_service_type': 'str',
        'metric_source': 'str',
        'name': 'str',
        'primary_owner_organization_id': 'int',
        'provider': 'str',
        'region_id': 'int',
        'suspended': 'bool',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'default_storage_id': 'default_storage_id',
        'edges': 'edges',
        'id': 'id',
        'immutable_slug': 'immutable_slug',
        'influx_bucket_name': 'influx_bucket_name',
        'is_custom_resource_spec_allowed': 'is_custom_resource_spec_allowed',
        'is_vessl_managed': 'is_vessl_managed',
        'kubernetes_container_runtime': 'kubernetes_container_runtime',
        'kubernetes_service_type': 'kubernetes_service_type',
        'metric_source': 'metric_source',
        'name': 'name',
        'primary_owner_organization_id': 'primary_owner_organization_id',
        'provider': 'provider',
        'region_id': 'region_id',
        'suspended': 'suspended',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, created_dt=None, default_storage_id=None, edges=None, id=None, immutable_slug=None, influx_bucket_name=None, is_custom_resource_spec_allowed=None, is_vessl_managed=None, kubernetes_container_runtime=None, kubernetes_service_type=None, metric_source=None, name=None, primary_owner_organization_id=None, provider=None, region_id=None, suspended=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """OrmKernelCluster - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._default_storage_id = None
        self._edges = None
        self._id = None
        self._immutable_slug = None
        self._influx_bucket_name = None
        self._is_custom_resource_spec_allowed = None
        self._is_vessl_managed = None
        self._kubernetes_container_runtime = None
        self._kubernetes_service_type = None
        self._metric_source = None
        self._name = None
        self._primary_owner_organization_id = None
        self._provider = None
        self._region_id = None
        self._suspended = None
        self._updated_dt = None
        self.discriminator = None

        if created_dt is not None:
            self.created_dt = created_dt
        if default_storage_id is not None:
            self.default_storage_id = default_storage_id
        if edges is not None:
            self.edges = edges
        if id is not None:
            self.id = id
        if immutable_slug is not None:
            self.immutable_slug = immutable_slug
        self.influx_bucket_name = influx_bucket_name
        if is_custom_resource_spec_allowed is not None:
            self.is_custom_resource_spec_allowed = is_custom_resource_spec_allowed
        if is_vessl_managed is not None:
            self.is_vessl_managed = is_vessl_managed
        if kubernetes_container_runtime is not None:
            self.kubernetes_container_runtime = kubernetes_container_runtime
        if kubernetes_service_type is not None:
            self.kubernetes_service_type = kubernetes_service_type
        self.metric_source = metric_source
        if name is not None:
            self.name = name
        if primary_owner_organization_id is not None:
            self.primary_owner_organization_id = primary_owner_organization_id
        if provider is not None:
            self.provider = provider
        if region_id is not None:
            self.region_id = region_id
        if suspended is not None:
            self.suspended = suspended
        if updated_dt is not None:
            self.updated_dt = updated_dt

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmKernelCluster.  # noqa: E501


        :return: The created_dt of this OrmKernelCluster.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmKernelCluster.


        :param created_dt: The created_dt of this OrmKernelCluster.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def default_storage_id(self):
        """Gets the default_storage_id of this OrmKernelCluster.  # noqa: E501


        :return: The default_storage_id of this OrmKernelCluster.  # noqa: E501
        :rtype: int
        """
        return self._default_storage_id

    @default_storage_id.setter
    def default_storage_id(self, default_storage_id):
        """Sets the default_storage_id of this OrmKernelCluster.


        :param default_storage_id: The default_storage_id of this OrmKernelCluster.  # noqa: E501
        :type default_storage_id: int
        """

        self._default_storage_id = default_storage_id

    @property
    def edges(self):
        """Gets the edges of this OrmKernelCluster.  # noqa: E501


        :return: The edges of this OrmKernelCluster.  # noqa: E501
        :rtype: OrmKernelClusterEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmKernelCluster.


        :param edges: The edges of this OrmKernelCluster.  # noqa: E501
        :type edges: OrmKernelClusterEdges
        """

        self._edges = edges

    @property
    def id(self):
        """Gets the id of this OrmKernelCluster.  # noqa: E501


        :return: The id of this OrmKernelCluster.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmKernelCluster.


        :param id: The id of this OrmKernelCluster.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def immutable_slug(self):
        """Gets the immutable_slug of this OrmKernelCluster.  # noqa: E501


        :return: The immutable_slug of this OrmKernelCluster.  # noqa: E501
        :rtype: str
        """
        return self._immutable_slug

    @immutable_slug.setter
    def immutable_slug(self, immutable_slug):
        """Sets the immutable_slug of this OrmKernelCluster.


        :param immutable_slug: The immutable_slug of this OrmKernelCluster.  # noqa: E501
        :type immutable_slug: str
        """

        self._immutable_slug = immutable_slug

    @property
    def influx_bucket_name(self):
        """Gets the influx_bucket_name of this OrmKernelCluster.  # noqa: E501


        :return: The influx_bucket_name of this OrmKernelCluster.  # noqa: E501
        :rtype: str
        """
        return self._influx_bucket_name

    @influx_bucket_name.setter
    def influx_bucket_name(self, influx_bucket_name):
        """Sets the influx_bucket_name of this OrmKernelCluster.


        :param influx_bucket_name: The influx_bucket_name of this OrmKernelCluster.  # noqa: E501
        :type influx_bucket_name: str
        """

        self._influx_bucket_name = influx_bucket_name

    @property
    def is_custom_resource_spec_allowed(self):
        """Gets the is_custom_resource_spec_allowed of this OrmKernelCluster.  # noqa: E501


        :return: The is_custom_resource_spec_allowed of this OrmKernelCluster.  # noqa: E501
        :rtype: bool
        """
        return self._is_custom_resource_spec_allowed

    @is_custom_resource_spec_allowed.setter
    def is_custom_resource_spec_allowed(self, is_custom_resource_spec_allowed):
        """Sets the is_custom_resource_spec_allowed of this OrmKernelCluster.


        :param is_custom_resource_spec_allowed: The is_custom_resource_spec_allowed of this OrmKernelCluster.  # noqa: E501
        :type is_custom_resource_spec_allowed: bool
        """

        self._is_custom_resource_spec_allowed = is_custom_resource_spec_allowed

    @property
    def is_vessl_managed(self):
        """Gets the is_vessl_managed of this OrmKernelCluster.  # noqa: E501


        :return: The is_vessl_managed of this OrmKernelCluster.  # noqa: E501
        :rtype: bool
        """
        return self._is_vessl_managed

    @is_vessl_managed.setter
    def is_vessl_managed(self, is_vessl_managed):
        """Sets the is_vessl_managed of this OrmKernelCluster.


        :param is_vessl_managed: The is_vessl_managed of this OrmKernelCluster.  # noqa: E501
        :type is_vessl_managed: bool
        """

        self._is_vessl_managed = is_vessl_managed

    @property
    def kubernetes_container_runtime(self):
        """Gets the kubernetes_container_runtime of this OrmKernelCluster.  # noqa: E501


        :return: The kubernetes_container_runtime of this OrmKernelCluster.  # noqa: E501
        :rtype: str
        """
        return self._kubernetes_container_runtime

    @kubernetes_container_runtime.setter
    def kubernetes_container_runtime(self, kubernetes_container_runtime):
        """Sets the kubernetes_container_runtime of this OrmKernelCluster.


        :param kubernetes_container_runtime: The kubernetes_container_runtime of this OrmKernelCluster.  # noqa: E501
        :type kubernetes_container_runtime: str
        """

        self._kubernetes_container_runtime = kubernetes_container_runtime

    @property
    def kubernetes_service_type(self):
        """Gets the kubernetes_service_type of this OrmKernelCluster.  # noqa: E501


        :return: The kubernetes_service_type of this OrmKernelCluster.  # noqa: E501
        :rtype: str
        """
        return self._kubernetes_service_type

    @kubernetes_service_type.setter
    def kubernetes_service_type(self, kubernetes_service_type):
        """Sets the kubernetes_service_type of this OrmKernelCluster.


        :param kubernetes_service_type: The kubernetes_service_type of this OrmKernelCluster.  # noqa: E501
        :type kubernetes_service_type: str
        """

        self._kubernetes_service_type = kubernetes_service_type

    @property
    def metric_source(self):
        """Gets the metric_source of this OrmKernelCluster.  # noqa: E501


        :return: The metric_source of this OrmKernelCluster.  # noqa: E501
        :rtype: str
        """
        return self._metric_source

    @metric_source.setter
    def metric_source(self, metric_source):
        """Sets the metric_source of this OrmKernelCluster.


        :param metric_source: The metric_source of this OrmKernelCluster.  # noqa: E501
        :type metric_source: str
        """

        self._metric_source = metric_source

    @property
    def name(self):
        """Gets the name of this OrmKernelCluster.  # noqa: E501


        :return: The name of this OrmKernelCluster.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this OrmKernelCluster.


        :param name: The name of this OrmKernelCluster.  # noqa: E501
        :type name: str
        """

        self._name = name

    @property
    def primary_owner_organization_id(self):
        """Gets the primary_owner_organization_id of this OrmKernelCluster.  # noqa: E501


        :return: The primary_owner_organization_id of this OrmKernelCluster.  # noqa: E501
        :rtype: int
        """
        return self._primary_owner_organization_id

    @primary_owner_organization_id.setter
    def primary_owner_organization_id(self, primary_owner_organization_id):
        """Sets the primary_owner_organization_id of this OrmKernelCluster.


        :param primary_owner_organization_id: The primary_owner_organization_id of this OrmKernelCluster.  # noqa: E501
        :type primary_owner_organization_id: int
        """

        self._primary_owner_organization_id = primary_owner_organization_id

    @property
    def provider(self):
        """Gets the provider of this OrmKernelCluster.  # noqa: E501


        :return: The provider of this OrmKernelCluster.  # noqa: E501
        :rtype: str
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this OrmKernelCluster.


        :param provider: The provider of this OrmKernelCluster.  # noqa: E501
        :type provider: str
        """

        self._provider = provider

    @property
    def region_id(self):
        """Gets the region_id of this OrmKernelCluster.  # noqa: E501


        :return: The region_id of this OrmKernelCluster.  # noqa: E501
        :rtype: int
        """
        return self._region_id

    @region_id.setter
    def region_id(self, region_id):
        """Sets the region_id of this OrmKernelCluster.


        :param region_id: The region_id of this OrmKernelCluster.  # noqa: E501
        :type region_id: int
        """

        self._region_id = region_id

    @property
    def suspended(self):
        """Gets the suspended of this OrmKernelCluster.  # noqa: E501


        :return: The suspended of this OrmKernelCluster.  # noqa: E501
        :rtype: bool
        """
        return self._suspended

    @suspended.setter
    def suspended(self, suspended):
        """Sets the suspended of this OrmKernelCluster.


        :param suspended: The suspended of this OrmKernelCluster.  # noqa: E501
        :type suspended: bool
        """

        self._suspended = suspended

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmKernelCluster.  # noqa: E501


        :return: The updated_dt of this OrmKernelCluster.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmKernelCluster.


        :param updated_dt: The updated_dt of this OrmKernelCluster.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmKernelCluster):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmKernelCluster):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmModelServiceRolloutStepWrapper(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'create_new_revision': 'OrmModelServiceRolloutStepCreateNewRevision',
        'send_notification': 'OrmModelServiceRolloutStepSendNotification',
        'step_type': 'str',
        'update_endpoint': 'OrmModelServiceRolloutStepUpdateEndpoint',
        'update_revisions': 'OrmModelServiceRolloutStepUpdateRevisions',
        'wait': 'OrmModelServiceRolloutStepWait'
    }

    attribute_map = {
        'create_new_revision': 'create_new_revision',
        'send_notification': 'send_notification',
        'step_type': 'step_type',
        'update_endpoint': 'update_endpoint',
        'update_revisions': 'update_revisions',
        'wait': 'wait'
    }

    def __init__(self, create_new_revision=None, send_notification=None, step_type=None, update_endpoint=None, update_revisions=None, wait=None, local_vars_configuration=None):  # noqa: E501
        """OrmModelServiceRolloutStepWrapper - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._create_new_revision = None
        self._send_notification = None
        self._step_type = None
        self._update_endpoint = None
        self._update_revisions = None
        self._wait = None
        self.discriminator = None

        if create_new_revision is not None:
            self.create_new_revision = create_new_revision
        if send_notification is not None:
            self.send_notification = send_notification
        if step_type is not None:
            self.step_type = step_type
        if update_endpoint is not None:
            self.update_endpoint = update_endpoint
        if update_revisions is not None:
            self.update_revisions = update_revisions
        if wait is not None:
            self.wait = wait

    @property
    def create_new_revision(self):
        """Gets the create_new_revision of this OrmModelServiceRolloutStepWrapper.  # noqa: E501


        :return: The create_new_revision of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :rtype: OrmModelServiceRolloutStepCreateNewRevision
        """
        return self._create_new_revision

    @create_new_revision.setter
    def create_new_revision(self, create_new_revision):
        """Sets the create_new_revision of this OrmModelServiceRolloutStepWrapper.


        :param create_new_revision: The create_new_revision of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :type create_new_revision: OrmModelServiceRolloutStepCreateNewRevision
        """

        self._create_new_revision = create_new_revision

    @property
    def send_notification(self):
        """Gets the send_notification of this OrmModelServiceRolloutStepWrapper.  # noqa: E501


        :return: The send_notification of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :rtype: OrmModelServiceRolloutStepSendNotification
        """
        return self._send_notification

    @send_notification.setter
    def send_notification(self, send_notification):
        """Sets the send_notification of this OrmModelServiceRolloutStepWrapper.


        :param send_notification: The send_notification of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :type send_notification: OrmModelServiceRolloutStepSendNotification
        """

        self._send_notification = send_notification

    @property
    def step_type(self):
        """Gets the step_type of this OrmModelServiceRolloutStepWrapper.  # noqa: E501


        :return: The step_type of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :rtype: str
        """
        return self._step_type

    @step_type.setter
    def step_type(self, step_type):
        """Sets the step_type of this OrmModelServiceRolloutStepWrapper.


        :param step_type: The step_type of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :type step_type: str
        """
        allowed_values = ["create_new_revision", "update_revisions", "update_endpoint", "wait", "send_notification"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and step_type not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `step_type` ({0}), must be one of {1}"  # noqa: E501
                .format(step_type, allowed_values)
            )

        self._step_type = step_type

    @property
    def update_endpoint(self):
        """Gets the update_endpoint of this OrmModelServiceRolloutStepWrapper.  # noqa: E501


        :return: The update_endpoint of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :rtype: OrmModelServiceRolloutStepUpdateEndpoint
        """
        return self._update_endpoint

    @update_endpoint.setter
    def update_endpoint(self, update_endpoint):
        """Sets the update_endpoint of this OrmModelServiceRolloutStepWrapper.


        :param update_endpoint: The update_endpoint of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :type update_endpoint: OrmModelServiceRolloutStepUpdateEndpoint
        """

        self._update_endpoint = update_endpoint

    @property
    def update_revisions(self):
        """Gets the update_revisions of this OrmModelServiceRolloutStepWrapper.  # noqa: E501


        :return: The update_revisions of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :rtype: OrmModelServiceRolloutStepUpdateRevisions
        """
        return self._update_revisions

    @update_revisions.setter
    def update_revisions(self, update_revisions):
        """Sets the update_revisions of this OrmModelServiceRolloutStepWrapper.


        :param update_revisions: The update_revisions of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :type update_revisions: OrmModelServiceRolloutStepUpdateRevisions
        """

        self._update_revisions = update_revisions

    @property
    def wait(self):
        """Gets the wait of this OrmModelServiceRolloutStepWrapper.  # noqa: E501


        :return: The wait of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :rtype: OrmModelServiceRolloutStepWait
        """
        return self._wait

    @wait.setter
    def wait(self, wait):
        """Sets the wait of this OrmModelServiceRolloutStepWrapper.


        :param wait: The wait of this OrmModelServiceRolloutStepWrapper.  # noqa: E501
        :type wait: OrmModelServiceRolloutStepWait
        """

        self._wait = wait

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmModelServiceRolloutStepWrapper):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmModelServiceRolloutStepWrapper):
            return True

        return self.to_dict() != other.to_dict()



import os
from typing import Optional, Tuple

from openapi_client.models import ResponseProjectInfo
from vessl import vessl_api
from vessl.util.exception import GitError
from vessl.util.git_local_repo import GitRepository
from vessl.util.uploader import Uploader


def get_git_repo() -> Tuple[Optional[str], Optional[str]]:
    if vessl_api.default_git_repo is None:
        raise GitError("Not in a git repository")

    owner, repo, _ = vessl_api.default_git_repo._get_github_repo()
    return owner, repo


def get_git_branch(**kwargs) -> Optional[str]:
    if vessl_api.default_git_repo is None:
        raise GitError("Not in a git repository")

    if "git_branch" in kwargs:
        return kwargs["git_branch"]
    if vessl_api.default_git_repo is not None:
        return vessl_api.default_git_repo.branch
    raise None


def get_git_ref(**kwargs) -> Optional[str]:
    if vessl_api.default_git_repo is None:
        raise GitError("Not in a git repository")

    git_ref = vessl_api.default_git_repo.commit_ref
    if "git_ref" in kwargs:
        git_ref = kwargs["git_ref"]

    if not vessl_api.default_git_repo.check_revision_in_remote(git_ref):
        raise GitError(f"Git commit does not exist in a remote repository: {git_ref}")

    return git_ref


def get_git_diff_path(project: ResponseProjectInfo, **kwargs) -> Optional[str]:
    if vessl_api.default_git_repo is None:
        raise GitError("Not in a git repository")

    use_git_diff = kwargs.get("use_git_diff", True)
    use_git_diff_untracked = kwargs.get("use_git_diff_untracked", True)

    if not use_git_diff:
        return None

    has_diff, _ = GitRepository.get_current_diff_status(revision_or_branch=get_git_ref())
    if not has_diff:
        return None

    local_file = GitRepository.get_current_diff_file(
        revision_or_branch=get_git_ref(), with_untracked=use_git_diff_untracked
    )

    remote_file = Uploader.upload(
        local_path=local_file.name,
        volume_id=project.volume_id,
        remote_path=os.path.basename(local_file.name),
    )

    local_file.close()
    return remote_file.path



import os
import sys
import threading
import time
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import psutil
import pynvml

from openapi_client.models import ExperimentMetricEntry
from vessl.util import logger
from vessl.util.common import safe_cast


class Collector(ABC):
    DROP_THRESHOLD = 360

    def __init__(self):
        self.buffer: List[ExperimentMetricEntry] = []
        self.lock = threading.Lock()

    def add(self, entries: List[ExperimentMetricEntry]):
        with self.lock:
            self.buffer += entries

    def collect(self):
        with self.lock:
            if len(self.buffer) > self.DROP_THRESHOLD:
                dropped = len(self.buffer) - self.DROP_THRESHOLD
                logger.warning(f"{dropped} collector entries dropped")
                self.buffer = self.buffer[-self.DROP_THRESHOLD :]

            return self.buffer[:]

    def truncate(self, idx):
        with self.lock:
            self.buffer = self.buffer[idx:]

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass


class IOCollector(Collector):
    def create_write_hook(self, orig_write, io_name):
        def _write(s: str) -> int:
            if s.strip() != "":
                self.add(
                    [
                        ExperimentMetricEntry(
                            measurement="log",
                            ts=time.time(),
                            tags={},
                            fields={io_name: s},
                        )
                    ]
                )
            return orig_write(s)

        return _write

    def __init__(self):
        super().__init__()

    def start(self):
        self.old_stdout_write = sys.stdout.write
        self.old_stderr_write = sys.stderr.write

        stdout_write_hook = self.create_write_hook(sys.stdout.write, "stdout")
        stderr_write_hook = self.create_write_hook(sys.stderr.write, "stderr")

        sys.stdout.write = stdout_write_hook
        sys.stderr.write = stderr_write_hook

    def create_callback(self, io_name):
        def f(ts, content):
            self.add(
                [
                    ExperimentMetricEntry(
                        measurement="log", ts=ts, tags={}, fields={io_name: content}
                    )
                ]
            )

        return f

    def stop(self):
        sys.stdout.write = self.old_stdout_write
        sys.stderr.write = self.old_stderr_write


class SystemMetricCollector(Collector):
    def __init__(self, gpu_count: int) -> None:
        super().__init__()
        self._gpu_count = gpu_count
        self._proc = psutil.Process()
        self._collect_interval_sec: float = 5
        self._thread = threading.Thread(target=self._thread_body, daemon=True)
        self._exit = threading.Event()

    def start(self):
        self._thread.start()

    def stop(self):
        self._exit.set()
        self._thread.join()

    def _thread_body(self):
        # Call to set latest value and junk return
        # https://psutil.readthedocs.io/en/latest/index.html?highlight=pcputimes#psutil.cpu_percent
        with self._proc.oneshot():
            psutil.cpu_percent()
            self._proc.cpu_percent()

        while not self._exit.is_set():
            self.add(entries=self._stats())
            self._exit.wait(timeout=self._collect_interval_sec)

    def _stats(self) -> List[ExperimentMetricEntry]:
        ts = time.time()
        entries = []
        with self._proc.oneshot():
            entries.append(
                ExperimentMetricEntry(
                    measurement="system_metric",
                    ts=ts,
                    fields={
                        "process_cpu_percent": self._proc.cpu_percent(),
                        "process_mem_percent": self._proc.memory_percent(),
                        "device_cpu_percent": psutil.cpu_percent(),
                        "device_mem_percent": psutil.virtual_memory().percent,
                    },
                    tags={
                        "type": "local",
                    },
                )
            )

        for i in range(self._gpu_count):
            try:
                handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            except pynvml.nvml.NVMLError:
                continue
            gpu_name = pynvml.nvmlDeviceGetName(handle).decode()
            component = f"{gpu_name} #{i+1}"
            use = pynvml.nvmlDeviceGetUtilizationRates(handle)

            entries.append(
                ExperimentMetricEntry(
                    measurement="system_metric",
                    ts=ts,
                    fields={
                        "device_gpu_util_percent": use.gpu,
                        "device_gpu_mem_percent": use.memory,
                    },
                    tags={
                        "type": "local",
                        "component": component,
                    },
                )
            )

        return entries


class UserMetricCollector(Collector):
    def __init__(self) -> None:
        super().__init__()
        self.step = 0

    def start(self):
        pass

    def stop(self):
        pass

    def handle_step(self, step: Optional[int]):
        if step is not None:
            if step < self.step:
                logger.warning("Step should not go backwards")
            self.step = step
        else:
            self.step += 1

    def _read_distributed_info(self) -> Dict[str, int]:
        distributed_info = {}
        distributed_number = safe_cast(os.environ.get("VESSL_DISTRIBUTED_NUMBER", None), int)
        if distributed_number is not None:
            distributed_info = {"distributed_number": distributed_number}
        else:
            workload_id = safe_cast(os.environ.get("VESSL_WORKLOAD_ID", None), int)
            if workload_id is not None:
                distributed_info = {"workload_id": workload_id}

        return distributed_info

    def build_metric_payload(self, payload: Dict[str, Any], ts: float):
        fields = payload
        tags = {"step": str(self.step)}

        distributed_info = self._read_distributed_info()
        if "distributed_number" in distributed_info:
            tags["distributed_number"] = str(distributed_info["distributed_number"])
        if "workload_id" in distributed_info:
            fields["workload_id"] = distributed_info["workload_id"]

        return ExperimentMetricEntry(
            measurement="experiment_plot_metric",
            ts=ts,
            tags=tags,
            fields=fields,
        )

    def build_media_payload(self, payload: Dict[str, Any], ts: float):
        fields = payload
        tags = {"step": str(self.step)}

        distributed_info = self._read_distributed_info()
        if "distributed_number" in distributed_info:
            tags["distributed_number"] = str(distributed_info["distributed_number"])
        if "workload_id" in distributed_info:
            fields["workload_id"] = distributed_info["workload_id"]

        return ExperimentMetricEntry(
            measurement="experiment_plot_file",
            ts=ts,
            tags=tags,
            fields=fields,
        )

    def log_metrics(self, payloads: List[ExperimentMetricEntry]) -> int:
        self.add(payloads)
        return self.step

    def log_media(self, payloads: List[ExperimentMetricEntry]) -> int:
        self.add(payloads)
        return self.step


class K8sCollector(Collector):
    def __init__(self) -> None:
        super().__init__()

    def start(self):
        # TODO
        pass

    def stop(self):
        # TODO
        pass



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmPipelineExecutionEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'pipeline': 'OrmPipeline',
        'spec': 'OrmPipelineSpec',
        'step_executions': 'list[OrmPipelineStepExecution]',
        'triggered_pipeline_step_execution': 'OrmPipelineStepExecution',
        'triggered_user': 'OrmUser'
    }

    attribute_map = {
        'pipeline': 'pipeline',
        'spec': 'spec',
        'step_executions': 'step_executions',
        'triggered_pipeline_step_execution': 'triggered_pipeline_step_execution',
        'triggered_user': 'triggered_user'
    }

    def __init__(self, pipeline=None, spec=None, step_executions=None, triggered_pipeline_step_execution=None, triggered_user=None, local_vars_configuration=None):  # noqa: E501
        """OrmPipelineExecutionEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._pipeline = None
        self._spec = None
        self._step_executions = None
        self._triggered_pipeline_step_execution = None
        self._triggered_user = None
        self.discriminator = None

        if pipeline is not None:
            self.pipeline = pipeline
        if spec is not None:
            self.spec = spec
        if step_executions is not None:
            self.step_executions = step_executions
        if triggered_pipeline_step_execution is not None:
            self.triggered_pipeline_step_execution = triggered_pipeline_step_execution
        if triggered_user is not None:
            self.triggered_user = triggered_user

    @property
    def pipeline(self):
        """Gets the pipeline of this OrmPipelineExecutionEdges.  # noqa: E501


        :return: The pipeline of this OrmPipelineExecutionEdges.  # noqa: E501
        :rtype: OrmPipeline
        """
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline):
        """Sets the pipeline of this OrmPipelineExecutionEdges.


        :param pipeline: The pipeline of this OrmPipelineExecutionEdges.  # noqa: E501
        :type pipeline: OrmPipeline
        """

        self._pipeline = pipeline

    @property
    def spec(self):
        """Gets the spec of this OrmPipelineExecutionEdges.  # noqa: E501


        :return: The spec of this OrmPipelineExecutionEdges.  # noqa: E501
        :rtype: OrmPipelineSpec
        """
        return self._spec

    @spec.setter
    def spec(self, spec):
        """Sets the spec of this OrmPipelineExecutionEdges.


        :param spec: The spec of this OrmPipelineExecutionEdges.  # noqa: E501
        :type spec: OrmPipelineSpec
        """

        self._spec = spec

    @property
    def step_executions(self):
        """Gets the step_executions of this OrmPipelineExecutionEdges.  # noqa: E501


        :return: The step_executions of this OrmPipelineExecutionEdges.  # noqa: E501
        :rtype: list[OrmPipelineStepExecution]
        """
        return self._step_executions

    @step_executions.setter
    def step_executions(self, step_executions):
        """Sets the step_executions of this OrmPipelineExecutionEdges.


        :param step_executions: The step_executions of this OrmPipelineExecutionEdges.  # noqa: E501
        :type step_executions: list[OrmPipelineStepExecution]
        """

        self._step_executions = step_executions

    @property
    def triggered_pipeline_step_execution(self):
        """Gets the triggered_pipeline_step_execution of this OrmPipelineExecutionEdges.  # noqa: E501


        :return: The triggered_pipeline_step_execution of this OrmPipelineExecutionEdges.  # noqa: E501
        :rtype: OrmPipelineStepExecution
        """
        return self._triggered_pipeline_step_execution

    @triggered_pipeline_step_execution.setter
    def triggered_pipeline_step_execution(self, triggered_pipeline_step_execution):
        """Sets the triggered_pipeline_step_execution of this OrmPipelineExecutionEdges.


        :param triggered_pipeline_step_execution: The triggered_pipeline_step_execution of this OrmPipelineExecutionEdges.  # noqa: E501
        :type triggered_pipeline_step_execution: OrmPipelineStepExecution
        """

        self._triggered_pipeline_step_execution = triggered_pipeline_step_execution

    @property
    def triggered_user(self):
        """Gets the triggered_user of this OrmPipelineExecutionEdges.  # noqa: E501


        :return: The triggered_user of this OrmPipelineExecutionEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._triggered_user

    @triggered_user.setter
    def triggered_user(self, triggered_user):
        """Sets the triggered_user of this OrmPipelineExecutionEdges.


        :param triggered_user: The triggered_user of this OrmPipelineExecutionEdges.  # noqa: E501
        :type triggered_user: OrmUser
        """

        self._triggered_user = triggered_user

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmPipelineExecutionEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmPipelineExecutionEdges):
            return True

        return self.to_dict() != other.to_dict()



import threading

from urllib3.exceptions import MaxRetryError

from openapi_client.models.experiment_progress_update_api_input import (
    ExperimentProgressUpdateAPIInput,
)
from vessl.util import logger
from vessl.util.api import VesslApi
from vessl.util.exception import VesslApiException

PROGRESS_UPDATE_INTERVAL_IN_SEC = 1


class ProgressUpdater(object):
    def __init__(self, api: VesslApi, experiment_id: int):
        self._api = api
        self._experiment_id = experiment_id
        self._thread = threading.Thread(target=self._thread_body, daemon=True)
        self._exit = threading.Event()

        self._progress = None

    def start(self):
        self._thread.start()

    def stop(self):
        self._exit.set()
        self._thread.join()

    def update(self, value: float):
        self._progress = value

    def _thread_body(self):
        while not self._exit.is_set():
            self._send()
            self._exit.wait(timeout=PROGRESS_UPDATE_INTERVAL_IN_SEC)
        self._send()

    def _send(self):
        if self._progress is None:
            return

        logger.debug(f"Sending experiment progress: {self._progress}")
        try:
            self._api.experiment_progress_update_api(
                self._experiment_id,
                experiment_progress_update_api_input=ExperimentProgressUpdateAPIInput(
                    progress_percent=self._progress,
                ),
            )
            self._progress = None  # Flush after sending

        except (MaxRetryError, VesslApiException) as e:
            logger.exception("Failed to send metrics to server", exc_info=e)

        except Exception as e:
            logger.exception("Unexpected error", exc_info=e)



from typing import List, Tuple

from openapi_client import ResponseSweepExperimentInfo
from openapi_client.models import (
    InfluxdbSweepLog,
    OrmEarlyStoppingSetting,
    OrmEarlyStoppingSpec,
    OrmHyperparameter,
)
from openapi_client.models import OrmParameter as SweepParameter
from openapi_client.models import OrmSweepObjective as SweepObjective
from openapi_client.models import (
    OrmSweepSearchSpace,
    ResponseSweepInfo,
    ResponseSweepListResponse,
    SweepCreateAPIInput,
)
from vessl import vessl_api
from vessl.kernel_cluster import read_cluster
from vessl.kernel_resource_spec import (
    _configure_custom_kernel_resource_spec,
    read_kernel_resource_spec,
)
from vessl.organization import _get_organization_name
from vessl.project import _get_project_name
from vessl.util.constant import MOUNT_PATH_OUTPUT, SWEEP_OBJECTIVE_TYPE_MAXIMIZE
from vessl.util.exception import InvalidTypeError
from vessl.volume import _configure_volume_mount_requests


def read_sweep(sweep_name: str, **kwargs) -> ResponseSweepInfo:
    """Read sweep in the default organization/project. If you want to
    override the default organization/project, then pass `organization_name` or
    `project_name` as `**kwargs`.

    Args:
        sweep_name(str): Sweep name.

    Example:
        ```python
        vessl.read_sweep(
            sweep_name="pitch-lord",
        )
        ```
    """
    return vessl_api.sweep_read_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        sweep=sweep_name,
    )


def list_sweeps(**kwargs) -> List[ResponseSweepListResponse]:
    """List sweeps in the default organization/project. If you want to
    override the default organization/project, then pass `organization_name` or
    `project_name` as `**kwargs`.

    Example:
        ```python
        vessl.list_sweeps()
        ```
    """
    return vessl_api.sweep_list_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
    ).results


def create_sweep(
    name: str,
    algorithm: str,
    parameters: List[SweepParameter],
    cluster_name: str,
    command: str,
    objective: SweepObjective = None,
    max_experiment_count: int = None,
    parallel_experiment_count: int = None,
    max_failed_experiment_count: int = None,
    resource_spec_name: str = None,
    processor_type: str = None,
    cpu_limit: float = None,
    memory_limit: str = None,
    gpu_type: str = "Any",
    gpu_limit: int = None,
    image_url: str = None,
    *,
    early_stopping_name: str = None,
    early_stopping_settings: List[Tuple[str, str]] = None,
    message: str = None,
    hyperparameters: List[Tuple[str, str]] = None,
    dataset_mounts: List[str] = None,
    git_ref_mounts: List[str] = None,
    git_diff_mount: str = None,
    archive_file_mount: str = None,
    object_storage_mount: str = None,
    root_volume_size: str = None,
    working_dir: str = None,
    output_dir: str = MOUNT_PATH_OUTPUT,
    **kwargs,
) -> ResponseSweepInfo:
    """Create sweep in the default organization/project. If you want to
    override the default organization/project, then pass `organization_name` or
    `project_name` as `**kwargs`. Pass `use_git_diff=True` if you want to run
    experiment with uncommitted changes and pass `use_git_diff_untracked=True`
    if you want to run untracked changes(only valid if `use_git_diff` is set).

    Args:
        name(str): Name
        objective(Optional[vessl.SweepObjective]): A sweep objective including goal, metric,
            and type.
        max_experiment_count(Optional[int]): The maximum number of experiments to run.
            Required unless grid search.
        parallel_experiment_count(Optional[int]): The number of experiments to run in
            parallel. Default: 1.
        max_failed_experiment_count(Optional[int]): The maximum number of experiments to
            allow to fail. Default: 1.
        algorithm(str): Parameter suggestion algorithm. `grid`, `random`, or
            `bayesian`.
        parameters(List[vessl.SweepParameter]): A list of parameters to search.
            - SweepParameter
                - name(str): The names of hyperparameters to search.
                - type(str): `int`, `double`, `categorical`.
                - range(SweepParameterRange): Search range.
                    - list(List[str]): A list of values to try.
                        If `list` is given, `min`, `max` and `step` will be ignored.
                    - min(str): The minimum value of the search range (inclusive).
                    - max(str): The maximum value of the search range (inclusive).
                    - step(Optional[str]): If provided, the values are limited to min + n*step.
        cluster_name(str): Cluster name(must be specified before other options).
        command(str): Start command to execute in experiment container.
        resource_spec_name(str):  Resource type to run an experiment (for
            managed cluster only). Defaults to None.
        processor_type(str) cpu or gpu (for custom cluster only). Defaults to
            None.
        cpu_limit(float): Number of vCPUs (for custom cluster only). Defaults to
            None.
        memory_limit(str): Memory limit (for custom cluster only).
            Defaults to None. Example: "100Gi", "500Mi"
        gpu_type(str): GPU type(name) (for custom cluster only). Defaults to "Any".
            Example: "RTX-3090-Ti" (Currently the exact name can be found on new experiment page)
        gpu_limit(int): Number of GPU cores (for custom cluster only). Defaults
            to None.
        image_url(str): Kernel docker image URL. Defaults to None.
        early_stopping_name(str): Early stopping algorithm name. Defaults to
            None.
        early_stopping_settings(List[Tuple[str, str]]): Early stopping algorithm
            settings. Defaults to None.
        message(str): Message. Defaults to None.
        hyperparameters(List[str]): A list of fixed hyperparameters. Defaults to None.
        dataset_mounts(List[str]): A list of dataset mounts. Defaults to None.
        git_ref_mounts(List[str]): A list of git repository mounts. Defaults to
            None.
        git_diff_mount(str): Git diff mounts. Defaults to None.
        archive_file_mount(str): Local archive file mounts. Defaults to None.
        object_storage_mount(str): Object storage mounts. Defaults to None.
        root_volume_size(str): Root volume size. Defaults to None.
        working_dir(str): Working directory path. Defaults to None.
        output_dir(str): Output directory path. Defaults to "/output/".

    Example:
        ```python
        sweep_objective = vessl.SweepObjective(
            type="maximize",
            goal="0.99",
            metric="val_accuracy",
        )

        parameters = [
            vessl.SweepParameter(
                name="optimizer",
                type="categorical",
                range=vessl.SweepParameterRange(
                    list=["adam", "sgd", "adadelta"]
                )
            ),
            vessl.SweepParameter(
                name="batch_size",
                type="int",
                range=vessl.SweepParameterRange(
                    max="256",
                    min="64",
                    step="8",
                )
            )
        ]

        # Custom Cluster
        vessl.create_sweep(
            name="example-sweep-name",
            objective=sweep_objective,
            max_experiment_count=4,
            parallel_experiment_count=2,
            max_failed_experiment_count=2,
            algorithm="random",
            parameters=parameters,
            dataset_mounts=["/input:mnist"],
            cluster_name="dgx-cluster",
            processor_type="gpu",
            gpu_type="NVIDIA-A100-SXM4-80GB",
            gpu_limit=2,
            cpu_limit=30,
            memory_limit="100Gi",
            kernel_image_url="public.ecr.aws/vessl/kernels:py36.full-cpu",
            start_command="pip install requirements.txt && python main.py",
        )

        # VESSL-Managed Cluster
        vessl.create_sweep(
            name="example-sweep-name",
            objective=sweep_objective,
            max_experiment_count=4,
            parallel_experiment_count=2,
            max_failed_experiment_count=2,
            algorithm="random",
            parameters=parameters,
            dataset_mounts=["/input:mnist"],
            cluster_name="aws-apne2",
            kernel_resource_spec_name="v1.cpu-4.mem-13",
            kernel_image_url="public.ecr.aws/vessl/kernels:py36.full-cpu",
            start_command="pip install requirements.txt && python main.py",
        )
        ```
    """
    cluster = read_cluster(cluster_name)

    kernel_resource_spec = kernel_resource_spec_id = None
    if resource_spec_name:
        kernel_resource_spec_id = read_kernel_resource_spec(
            cluster.id,
            resource_spec_name,
        ).id
    else:
        kernel_resource_spec = _configure_custom_kernel_resource_spec(
            processor_type,
            cpu_limit,
            memory_limit,
            gpu_type,
            gpu_limit,
        )

    early_stopping_spec = OrmEarlyStoppingSpec(
        algorithm_name=early_stopping_name,
        algorithm_settings=[
            OrmEarlyStoppingSetting(name=k, value=str(v))
            for k, v in (early_stopping_settings or [])
        ],
    )

    volume_mount_requests = _configure_volume_mount_requests(
        dataset_mounts=dataset_mounts,
        git_ref_mounts=git_ref_mounts,
        git_diff_mount=git_diff_mount,
        archive_file_mount=archive_file_mount,
        object_storage_mounts=object_storage_mount,
        root_volume_size=root_volume_size,
        working_dir=working_dir,
        output_dir=output_dir,
        **kwargs,
    )

    return vessl_api.sweep_create_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        sweep_create_api_input=SweepCreateAPIInput(
            name=name,
            objective=objective,
            max_experiment_count=max_experiment_count,
            parallel_experiment_count=parallel_experiment_count,
            max_failed_experiment_count=max_failed_experiment_count,
            algorithm=algorithm,
            search_space=OrmSweepSearchSpace(
                parameters=parameters,
            ),
            cluster_id=cluster.id,
            hyperparameters=[
                OrmHyperparameter(key=key, value=str(value))
                for key, value in map(
                    lambda hyperparameter: hyperparameter.split("="),
                    (hyperparameters or []),
                )
            ],
            image_url=image_url,
            early_stopping_spec=early_stopping_spec,
            message=message,
            resource_spec=kernel_resource_spec,
            resource_spec_id=kernel_resource_spec_id,
            start_command=command,
            volumes=volume_mount_requests,
        ),
    )


def terminate_sweep(sweep_name: str, **kwargs) -> ResponseSweepInfo:
    """Terminate sweep in the default organization/project. If you want to
    override the default organization/project, then pass `organization_name` or
    `project_name` as `**kwargs`.

    Args:
        sweep_name(str): Sweep name.

    Example:
        ```python
        vessl.terminate_sweep(
            sweep_name="pitch-lord",
        )
        ```
    """
    return vessl_api.sweep_terminate_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        sweep=sweep_name,
    )


def list_sweep_logs(sweep_name: str, tail: int = 200, **kwargs) -> List[InfluxdbSweepLog]:
    """List sweep logs in the default organization/project. If you want to
    override the default organization/project, then pass `organization_name` or
    `project_name` as `**kwargs`.

    Args:
        sweep_name(str): Sweep name.
        tail(int): The number of lines to display from the end. Display all if
            -1. Defaults to 200.

    Example:
        ```python
        vessl.list_sweep_logs(
            sweep_name="pitch-lord",
        )
        ```
    """

    if tail == -1:
        tail = None

    return vessl_api.sweep_logs_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        sweep=sweep_name,
        limit=tail,
    ).logs


def get_best_sweep_experiment(sweep_name: str, **kwargs) -> ResponseSweepExperimentInfo:
    """Read sweep and return the best experiment info in the default
    organization/project. If you want to override the default
    organization/project, then pass `organization_name` or `project_name` as
    `**kwargs`.

    Args:
        sweep_name(str): Sweep name.

    Example:
        ```python
        vessl.get_best_sweep_experiment(
            sweep_name="pitch-lord",
        )
        ```
    """

    sweep_read_response = vessl_api.sweep_read_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        sweep=sweep_name,
    )
    if type(sweep_read_response) != ResponseSweepInfo:
        return sweep_read_response

    def best_experiments(
        items: [ResponseSweepExperimentInfo], maximize: bool
    ) -> ResponseSweepExperimentInfo:
        return sorted(items, key=lambda x: x.objective_metric_value, reverse=maximize).pop()

    if sweep_read_response.experiment_summary.total > 0:
        filtered_experiment_summary = []
        for experiment in sweep_read_response.experiment_summary.sweep_experiment_infos:
            if not experiment.objective_metric_value:
                continue
            elif not isinstance(experiment.objective_metric_value, (int, float)):
                raise InvalidTypeError(
                    f"The type of experiment objective metric value is "
                    f"not integer or float: "
                    f"{type(experiment.objective_metric_value)}"
                )
            filtered_experiment_summary.append(experiment)

        if sweep_read_response.objective.type == SWEEP_OBJECTIVE_TYPE_MAXIMIZE:
            return best_experiments(filtered_experiment_summary, True)
        else:
            return best_experiments(filtered_experiment_summary, False)

    return ResponseSweepExperimentInfo()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseCodeRefsV2(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'git_ref': 'str',
        'mount_name': 'str',
        'mount_path': 'str',
        'protocol': 'str',
        'ref_http': 'ResponseCodeRefHTTP',
        'ref_ssh': 'ResponseCodeRefSSH'
    }

    attribute_map = {
        'git_ref': 'git_ref',
        'mount_name': 'mount_name',
        'mount_path': 'mount_path',
        'protocol': 'protocol',
        'ref_http': 'ref_http',
        'ref_ssh': 'ref_ssh'
    }

    def __init__(self, git_ref=None, mount_name=None, mount_path=None, protocol=None, ref_http=None, ref_ssh=None, local_vars_configuration=None):  # noqa: E501
        """ResponseCodeRefsV2 - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._git_ref = None
        self._mount_name = None
        self._mount_path = None
        self._protocol = None
        self._ref_http = None
        self._ref_ssh = None
        self.discriminator = None

        if git_ref is not None:
            self.git_ref = git_ref
        if mount_name is not None:
            self.mount_name = mount_name
        if mount_path is not None:
            self.mount_path = mount_path
        if protocol is not None:
            self.protocol = protocol
        if ref_http is not None:
            self.ref_http = ref_http
        if ref_ssh is not None:
            self.ref_ssh = ref_ssh

    @property
    def git_ref(self):
        """Gets the git_ref of this ResponseCodeRefsV2.  # noqa: E501


        :return: The git_ref of this ResponseCodeRefsV2.  # noqa: E501
        :rtype: str
        """
        return self._git_ref

    @git_ref.setter
    def git_ref(self, git_ref):
        """Sets the git_ref of this ResponseCodeRefsV2.


        :param git_ref: The git_ref of this ResponseCodeRefsV2.  # noqa: E501
        :type git_ref: str
        """

        self._git_ref = git_ref

    @property
    def mount_name(self):
        """Gets the mount_name of this ResponseCodeRefsV2.  # noqa: E501


        :return: The mount_name of this ResponseCodeRefsV2.  # noqa: E501
        :rtype: str
        """
        return self._mount_name

    @mount_name.setter
    def mount_name(self, mount_name):
        """Sets the mount_name of this ResponseCodeRefsV2.


        :param mount_name: The mount_name of this ResponseCodeRefsV2.  # noqa: E501
        :type mount_name: str
        """

        self._mount_name = mount_name

    @property
    def mount_path(self):
        """Gets the mount_path of this ResponseCodeRefsV2.  # noqa: E501


        :return: The mount_path of this ResponseCodeRefsV2.  # noqa: E501
        :rtype: str
        """
        return self._mount_path

    @mount_path.setter
    def mount_path(self, mount_path):
        """Sets the mount_path of this ResponseCodeRefsV2.


        :param mount_path: The mount_path of this ResponseCodeRefsV2.  # noqa: E501
        :type mount_path: str
        """

        self._mount_path = mount_path

    @property
    def protocol(self):
        """Gets the protocol of this ResponseCodeRefsV2.  # noqa: E501


        :return: The protocol of this ResponseCodeRefsV2.  # noqa: E501
        :rtype: str
        """
        return self._protocol

    @protocol.setter
    def protocol(self, protocol):
        """Sets the protocol of this ResponseCodeRefsV2.


        :param protocol: The protocol of this ResponseCodeRefsV2.  # noqa: E501
        :type protocol: str
        """
        allowed_values = ["http", "ssh"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and protocol not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `protocol` ({0}), must be one of {1}"  # noqa: E501
                .format(protocol, allowed_values)
            )

        self._protocol = protocol

    @property
    def ref_http(self):
        """Gets the ref_http of this ResponseCodeRefsV2.  # noqa: E501


        :return: The ref_http of this ResponseCodeRefsV2.  # noqa: E501
        :rtype: ResponseCodeRefHTTP
        """
        return self._ref_http

    @ref_http.setter
    def ref_http(self, ref_http):
        """Sets the ref_http of this ResponseCodeRefsV2.


        :param ref_http: The ref_http of this ResponseCodeRefsV2.  # noqa: E501
        :type ref_http: ResponseCodeRefHTTP
        """

        self._ref_http = ref_http

    @property
    def ref_ssh(self):
        """Gets the ref_ssh of this ResponseCodeRefsV2.  # noqa: E501


        :return: The ref_ssh of this ResponseCodeRefsV2.  # noqa: E501
        :rtype: ResponseCodeRefSSH
        """
        return self._ref_ssh

    @ref_ssh.setter
    def ref_ssh(self, ref_ssh):
        """Sets the ref_ssh of this ResponseCodeRefsV2.


        :param ref_ssh: The ref_ssh of this ResponseCodeRefsV2.  # noqa: E501
        :type ref_ssh: ResponseCodeRefSSH
        """

        self._ref_ssh = ref_ssh

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseCodeRefsV2):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseCodeRefsV2):
            return True

        return self.to_dict() != other.to_dict()



import sys

import click

import vessl
from vessl.cli._base import VesslGroup, vessl_argument
from vessl.cli._util import (
    Endpoint,
    format_url,
    generic_prompter,
    print_data,
    print_error,
    print_info,
    print_success,
    print_table,
    prompt_choices,
)
from vessl.cli.organization import organization_name_option
from vessl.project import create_project, list_projects, read_project
from vessl.util.exception import InvalidProjectError


def project_name_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
) -> str:
    projects = list_projects()
    return prompt_choices("Project", [x.name for x in projects])


@click.command(name="project", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=project_name_prompter)
@organization_name_option
def read(name: str):
    project = read_project(project_name=name)
    print_data(
        {
            "ID": project.id,
            "Name": project.name,
            "Experiments": project.experiment_summary.total,
        }
    )


@cli.vessl_command()
@organization_name_option
def list():
    projects = list_projects()
    print_table(
        projects,
        ["Name", "Experiments"],
        lambda x: [x.name, x.experiment_summary.total],
    )


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=generic_prompter("Project name"))
@click.option("-m", "--description", type=click.STRING)
@organization_name_option
def create(name: str, description: str):
    project = create_project(
        project_name=name,
        description=description,
    )
    print_success(
        f"Created '{project.name}'.\n"
        f"For more info: {format_url(Endpoint.project.format(project.organization.name, project.name))}"
    )


### Deprecate for now
# @cli.vessl_command()
# @vessl_argument(
#     "name", type=click.STRING, required=True, prompter=version_control_project_name_prompter,
# )
# @organization_name_option
# def clone(name: str):
#     clone_project(project_name=name)


def project_name_callback(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
):
    if vessl.vessl_api.organization is None:
        vessl.vessl_api.set_organization()

    try:
        project_name = vessl.vessl_api.set_project(value)
        print_info(f"Project: {project_name}")
    except InvalidProjectError:
        print_error("Invalid project. Please choose a project using `vessl configure`.")
        sys.exit(1)


# Ensure this is called before other options with `is_eager=True` for
# other callbacks that need organization to be preconfigured.
project_name_option = click.option(
    "--project",
    "project_name",
    type=click.STRING,
    hidden=True,
    is_eager=True,
    expose_value=False,
    callback=project_name_callback,
    help="Override default project.",
)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrganizationOrganizationBillingReadResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'credit_balance': 'float',
        'discounted_on_demand_amount': 'float',
        'discounted_on_premise_amount': 'float',
        'discounted_seats_amount': 'float',
        'invoice_status': 'str',
        'invoice_url': 'str',
        'on_demand_amount': 'float',
        'on_demand_discount_rate': 'float',
        'on_premise_amount': 'float',
        'on_premise_discount_rate': 'float',
        'on_premise_node_count': 'int',
        'payment_card_last_digits': 'str',
        'seats_amount': 'float',
        'seats_discount_rate': 'float'
    }

    attribute_map = {
        'credit_balance': 'credit_balance',
        'discounted_on_demand_amount': 'discounted_on_demand_amount',
        'discounted_on_premise_amount': 'discounted_on_premise_amount',
        'discounted_seats_amount': 'discounted_seats_amount',
        'invoice_status': 'invoice_status',
        'invoice_url': 'invoice_url',
        'on_demand_amount': 'on_demand_amount',
        'on_demand_discount_rate': 'on_demand_discount_rate',
        'on_premise_amount': 'on_premise_amount',
        'on_premise_discount_rate': 'on_premise_discount_rate',
        'on_premise_node_count': 'on_premise_node_count',
        'payment_card_last_digits': 'payment_card_last_digits',
        'seats_amount': 'seats_amount',
        'seats_discount_rate': 'seats_discount_rate'
    }

    def __init__(self, credit_balance=None, discounted_on_demand_amount=None, discounted_on_premise_amount=None, discounted_seats_amount=None, invoice_status=None, invoice_url=None, on_demand_amount=None, on_demand_discount_rate=None, on_premise_amount=None, on_premise_discount_rate=None, on_premise_node_count=None, payment_card_last_digits=None, seats_amount=None, seats_discount_rate=None, local_vars_configuration=None):  # noqa: E501
        """OrganizationOrganizationBillingReadResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._credit_balance = None
        self._discounted_on_demand_amount = None
        self._discounted_on_premise_amount = None
        self._discounted_seats_amount = None
        self._invoice_status = None
        self._invoice_url = None
        self._on_demand_amount = None
        self._on_demand_discount_rate = None
        self._on_premise_amount = None
        self._on_premise_discount_rate = None
        self._on_premise_node_count = None
        self._payment_card_last_digits = None
        self._seats_amount = None
        self._seats_discount_rate = None
        self.discriminator = None

        self.credit_balance = credit_balance
        self.discounted_on_demand_amount = discounted_on_demand_amount
        self.discounted_on_premise_amount = discounted_on_premise_amount
        self.discounted_seats_amount = discounted_seats_amount
        self.invoice_status = invoice_status
        self.invoice_url = invoice_url
        self.on_demand_amount = on_demand_amount
        self.on_demand_discount_rate = on_demand_discount_rate
        self.on_premise_amount = on_premise_amount
        self.on_premise_discount_rate = on_premise_discount_rate
        self.on_premise_node_count = on_premise_node_count
        self.payment_card_last_digits = payment_card_last_digits
        self.seats_amount = seats_amount
        self.seats_discount_rate = seats_discount_rate

    @property
    def credit_balance(self):
        """Gets the credit_balance of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The credit_balance of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._credit_balance

    @credit_balance.setter
    def credit_balance(self, credit_balance):
        """Sets the credit_balance of this OrganizationOrganizationBillingReadResponse.


        :param credit_balance: The credit_balance of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type credit_balance: float
        """
        if self.local_vars_configuration.client_side_validation and credit_balance is None:  # noqa: E501
            raise ValueError("Invalid value for `credit_balance`, must not be `None`")  # noqa: E501

        self._credit_balance = credit_balance

    @property
    def discounted_on_demand_amount(self):
        """Gets the discounted_on_demand_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The discounted_on_demand_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._discounted_on_demand_amount

    @discounted_on_demand_amount.setter
    def discounted_on_demand_amount(self, discounted_on_demand_amount):
        """Sets the discounted_on_demand_amount of this OrganizationOrganizationBillingReadResponse.


        :param discounted_on_demand_amount: The discounted_on_demand_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type discounted_on_demand_amount: float
        """
        if self.local_vars_configuration.client_side_validation and discounted_on_demand_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `discounted_on_demand_amount`, must not be `None`")  # noqa: E501

        self._discounted_on_demand_amount = discounted_on_demand_amount

    @property
    def discounted_on_premise_amount(self):
        """Gets the discounted_on_premise_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The discounted_on_premise_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._discounted_on_premise_amount

    @discounted_on_premise_amount.setter
    def discounted_on_premise_amount(self, discounted_on_premise_amount):
        """Sets the discounted_on_premise_amount of this OrganizationOrganizationBillingReadResponse.


        :param discounted_on_premise_amount: The discounted_on_premise_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type discounted_on_premise_amount: float
        """
        if self.local_vars_configuration.client_side_validation and discounted_on_premise_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `discounted_on_premise_amount`, must not be `None`")  # noqa: E501

        self._discounted_on_premise_amount = discounted_on_premise_amount

    @property
    def discounted_seats_amount(self):
        """Gets the discounted_seats_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The discounted_seats_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._discounted_seats_amount

    @discounted_seats_amount.setter
    def discounted_seats_amount(self, discounted_seats_amount):
        """Sets the discounted_seats_amount of this OrganizationOrganizationBillingReadResponse.


        :param discounted_seats_amount: The discounted_seats_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type discounted_seats_amount: float
        """
        if self.local_vars_configuration.client_side_validation and discounted_seats_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `discounted_seats_amount`, must not be `None`")  # noqa: E501

        self._discounted_seats_amount = discounted_seats_amount

    @property
    def invoice_status(self):
        """Gets the invoice_status of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The invoice_status of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: str
        """
        return self._invoice_status

    @invoice_status.setter
    def invoice_status(self, invoice_status):
        """Sets the invoice_status of this OrganizationOrganizationBillingReadResponse.


        :param invoice_status: The invoice_status of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type invoice_status: str
        """
        if self.local_vars_configuration.client_side_validation and invoice_status is None:  # noqa: E501
            raise ValueError("Invalid value for `invoice_status`, must not be `None`")  # noqa: E501

        self._invoice_status = invoice_status

    @property
    def invoice_url(self):
        """Gets the invoice_url of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The invoice_url of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: str
        """
        return self._invoice_url

    @invoice_url.setter
    def invoice_url(self, invoice_url):
        """Sets the invoice_url of this OrganizationOrganizationBillingReadResponse.


        :param invoice_url: The invoice_url of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type invoice_url: str
        """
        if self.local_vars_configuration.client_side_validation and invoice_url is None:  # noqa: E501
            raise ValueError("Invalid value for `invoice_url`, must not be `None`")  # noqa: E501

        self._invoice_url = invoice_url

    @property
    def on_demand_amount(self):
        """Gets the on_demand_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The on_demand_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._on_demand_amount

    @on_demand_amount.setter
    def on_demand_amount(self, on_demand_amount):
        """Sets the on_demand_amount of this OrganizationOrganizationBillingReadResponse.


        :param on_demand_amount: The on_demand_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type on_demand_amount: float
        """
        if self.local_vars_configuration.client_side_validation and on_demand_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `on_demand_amount`, must not be `None`")  # noqa: E501

        self._on_demand_amount = on_demand_amount

    @property
    def on_demand_discount_rate(self):
        """Gets the on_demand_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The on_demand_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._on_demand_discount_rate

    @on_demand_discount_rate.setter
    def on_demand_discount_rate(self, on_demand_discount_rate):
        """Sets the on_demand_discount_rate of this OrganizationOrganizationBillingReadResponse.


        :param on_demand_discount_rate: The on_demand_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type on_demand_discount_rate: float
        """
        if self.local_vars_configuration.client_side_validation and on_demand_discount_rate is None:  # noqa: E501
            raise ValueError("Invalid value for `on_demand_discount_rate`, must not be `None`")  # noqa: E501

        self._on_demand_discount_rate = on_demand_discount_rate

    @property
    def on_premise_amount(self):
        """Gets the on_premise_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The on_premise_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._on_premise_amount

    @on_premise_amount.setter
    def on_premise_amount(self, on_premise_amount):
        """Sets the on_premise_amount of this OrganizationOrganizationBillingReadResponse.


        :param on_premise_amount: The on_premise_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type on_premise_amount: float
        """
        if self.local_vars_configuration.client_side_validation and on_premise_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `on_premise_amount`, must not be `None`")  # noqa: E501

        self._on_premise_amount = on_premise_amount

    @property
    def on_premise_discount_rate(self):
        """Gets the on_premise_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The on_premise_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._on_premise_discount_rate

    @on_premise_discount_rate.setter
    def on_premise_discount_rate(self, on_premise_discount_rate):
        """Sets the on_premise_discount_rate of this OrganizationOrganizationBillingReadResponse.


        :param on_premise_discount_rate: The on_premise_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type on_premise_discount_rate: float
        """
        if self.local_vars_configuration.client_side_validation and on_premise_discount_rate is None:  # noqa: E501
            raise ValueError("Invalid value for `on_premise_discount_rate`, must not be `None`")  # noqa: E501

        self._on_premise_discount_rate = on_premise_discount_rate

    @property
    def on_premise_node_count(self):
        """Gets the on_premise_node_count of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The on_premise_node_count of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: int
        """
        return self._on_premise_node_count

    @on_premise_node_count.setter
    def on_premise_node_count(self, on_premise_node_count):
        """Sets the on_premise_node_count of this OrganizationOrganizationBillingReadResponse.


        :param on_premise_node_count: The on_premise_node_count of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type on_premise_node_count: int
        """
        if self.local_vars_configuration.client_side_validation and on_premise_node_count is None:  # noqa: E501
            raise ValueError("Invalid value for `on_premise_node_count`, must not be `None`")  # noqa: E501

        self._on_premise_node_count = on_premise_node_count

    @property
    def payment_card_last_digits(self):
        """Gets the payment_card_last_digits of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The payment_card_last_digits of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: str
        """
        return self._payment_card_last_digits

    @payment_card_last_digits.setter
    def payment_card_last_digits(self, payment_card_last_digits):
        """Sets the payment_card_last_digits of this OrganizationOrganizationBillingReadResponse.


        :param payment_card_last_digits: The payment_card_last_digits of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type payment_card_last_digits: str
        """
        if self.local_vars_configuration.client_side_validation and payment_card_last_digits is None:  # noqa: E501
            raise ValueError("Invalid value for `payment_card_last_digits`, must not be `None`")  # noqa: E501

        self._payment_card_last_digits = payment_card_last_digits

    @property
    def seats_amount(self):
        """Gets the seats_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The seats_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._seats_amount

    @seats_amount.setter
    def seats_amount(self, seats_amount):
        """Sets the seats_amount of this OrganizationOrganizationBillingReadResponse.


        :param seats_amount: The seats_amount of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type seats_amount: float
        """
        if self.local_vars_configuration.client_side_validation and seats_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `seats_amount`, must not be `None`")  # noqa: E501

        self._seats_amount = seats_amount

    @property
    def seats_discount_rate(self):
        """Gets the seats_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501


        :return: The seats_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :rtype: float
        """
        return self._seats_discount_rate

    @seats_discount_rate.setter
    def seats_discount_rate(self, seats_discount_rate):
        """Sets the seats_discount_rate of this OrganizationOrganizationBillingReadResponse.


        :param seats_discount_rate: The seats_discount_rate of this OrganizationOrganizationBillingReadResponse.  # noqa: E501
        :type seats_discount_rate: float
        """
        if self.local_vars_configuration.client_side_validation and seats_discount_rate is None:  # noqa: E501
            raise ValueError("Invalid value for `seats_discount_rate`, must not be `None`")  # noqa: E501

        self._seats_discount_rate = seats_discount_rate

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrganizationOrganizationBillingReadResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrganizationOrganizationBillingReadResponse):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseOrganization(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'access_control_config': 'OrmAccessControlConfig',
        'auto_top_up_charge_amount': 'float',
        'auto_top_up_enabled': 'bool',
        'auto_top_up_trigger_amount': 'float',
        'aws_external_id': 'str',
        'created_dt': 'datetime',
        'credit_balance': 'float',
        'description': 'str',
        'display_name': 'str',
        'id': 'int',
        'name': 'str',
        'owner_id': 'int',
        'pricing_plan': 'ResponsePricingPlan',
        'primary_owner_id': 'int',
        'show_tutorial': 'bool',
        'stripe_customer_id': 'str',
        'stripe_subscription_id': 'str',
        'tutorials': 'list[ResponseTutorialResponse]',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'access_control_config': 'access_control_config',
        'auto_top_up_charge_amount': 'auto_top_up_charge_amount',
        'auto_top_up_enabled': 'auto_top_up_enabled',
        'auto_top_up_trigger_amount': 'auto_top_up_trigger_amount',
        'aws_external_id': 'aws_external_id',
        'created_dt': 'created_dt',
        'credit_balance': 'credit_balance',
        'description': 'description',
        'display_name': 'display_name',
        'id': 'id',
        'name': 'name',
        'owner_id': 'owner_id',
        'pricing_plan': 'pricing_plan',
        'primary_owner_id': 'primary_owner_id',
        'show_tutorial': 'show_tutorial',
        'stripe_customer_id': 'stripe_customer_id',
        'stripe_subscription_id': 'stripe_subscription_id',
        'tutorials': 'tutorials',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, access_control_config=None, auto_top_up_charge_amount=None, auto_top_up_enabled=None, auto_top_up_trigger_amount=None, aws_external_id=None, created_dt=None, credit_balance=None, description=None, display_name=None, id=None, name=None, owner_id=None, pricing_plan=None, primary_owner_id=None, show_tutorial=None, stripe_customer_id=None, stripe_subscription_id=None, tutorials=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """ResponseOrganization - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._access_control_config = None
        self._auto_top_up_charge_amount = None
        self._auto_top_up_enabled = None
        self._auto_top_up_trigger_amount = None
        self._aws_external_id = None
        self._created_dt = None
        self._credit_balance = None
        self._description = None
        self._display_name = None
        self._id = None
        self._name = None
        self._owner_id = None
        self._pricing_plan = None
        self._primary_owner_id = None
        self._show_tutorial = None
        self._stripe_customer_id = None
        self._stripe_subscription_id = None
        self._tutorials = None
        self._updated_dt = None
        self.discriminator = None

        self.access_control_config = access_control_config
        self.auto_top_up_charge_amount = auto_top_up_charge_amount
        self.auto_top_up_enabled = auto_top_up_enabled
        self.auto_top_up_trigger_amount = auto_top_up_trigger_amount
        self.aws_external_id = aws_external_id
        self.created_dt = created_dt
        self.credit_balance = credit_balance
        self.description = description
        self.display_name = display_name
        self.id = id
        self.name = name
        self.owner_id = owner_id
        self.pricing_plan = pricing_plan
        self.primary_owner_id = primary_owner_id
        self.show_tutorial = show_tutorial
        self.stripe_customer_id = stripe_customer_id
        self.stripe_subscription_id = stripe_subscription_id
        self.tutorials = tutorials
        self.updated_dt = updated_dt

    @property
    def access_control_config(self):
        """Gets the access_control_config of this ResponseOrganization.  # noqa: E501


        :return: The access_control_config of this ResponseOrganization.  # noqa: E501
        :rtype: OrmAccessControlConfig
        """
        return self._access_control_config

    @access_control_config.setter
    def access_control_config(self, access_control_config):
        """Sets the access_control_config of this ResponseOrganization.


        :param access_control_config: The access_control_config of this ResponseOrganization.  # noqa: E501
        :type access_control_config: OrmAccessControlConfig
        """
        if self.local_vars_configuration.client_side_validation and access_control_config is None:  # noqa: E501
            raise ValueError("Invalid value for `access_control_config`, must not be `None`")  # noqa: E501

        self._access_control_config = access_control_config

    @property
    def auto_top_up_charge_amount(self):
        """Gets the auto_top_up_charge_amount of this ResponseOrganization.  # noqa: E501


        :return: The auto_top_up_charge_amount of this ResponseOrganization.  # noqa: E501
        :rtype: float
        """
        return self._auto_top_up_charge_amount

    @auto_top_up_charge_amount.setter
    def auto_top_up_charge_amount(self, auto_top_up_charge_amount):
        """Sets the auto_top_up_charge_amount of this ResponseOrganization.


        :param auto_top_up_charge_amount: The auto_top_up_charge_amount of this ResponseOrganization.  # noqa: E501
        :type auto_top_up_charge_amount: float
        """
        if self.local_vars_configuration.client_side_validation and auto_top_up_charge_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `auto_top_up_charge_amount`, must not be `None`")  # noqa: E501

        self._auto_top_up_charge_amount = auto_top_up_charge_amount

    @property
    def auto_top_up_enabled(self):
        """Gets the auto_top_up_enabled of this ResponseOrganization.  # noqa: E501


        :return: The auto_top_up_enabled of this ResponseOrganization.  # noqa: E501
        :rtype: bool
        """
        return self._auto_top_up_enabled

    @auto_top_up_enabled.setter
    def auto_top_up_enabled(self, auto_top_up_enabled):
        """Sets the auto_top_up_enabled of this ResponseOrganization.


        :param auto_top_up_enabled: The auto_top_up_enabled of this ResponseOrganization.  # noqa: E501
        :type auto_top_up_enabled: bool
        """
        if self.local_vars_configuration.client_side_validation and auto_top_up_enabled is None:  # noqa: E501
            raise ValueError("Invalid value for `auto_top_up_enabled`, must not be `None`")  # noqa: E501

        self._auto_top_up_enabled = auto_top_up_enabled

    @property
    def auto_top_up_trigger_amount(self):
        """Gets the auto_top_up_trigger_amount of this ResponseOrganization.  # noqa: E501


        :return: The auto_top_up_trigger_amount of this ResponseOrganization.  # noqa: E501
        :rtype: float
        """
        return self._auto_top_up_trigger_amount

    @auto_top_up_trigger_amount.setter
    def auto_top_up_trigger_amount(self, auto_top_up_trigger_amount):
        """Sets the auto_top_up_trigger_amount of this ResponseOrganization.


        :param auto_top_up_trigger_amount: The auto_top_up_trigger_amount of this ResponseOrganization.  # noqa: E501
        :type auto_top_up_trigger_amount: float
        """
        if self.local_vars_configuration.client_side_validation and auto_top_up_trigger_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `auto_top_up_trigger_amount`, must not be `None`")  # noqa: E501

        self._auto_top_up_trigger_amount = auto_top_up_trigger_amount

    @property
    def aws_external_id(self):
        """Gets the aws_external_id of this ResponseOrganization.  # noqa: E501


        :return: The aws_external_id of this ResponseOrganization.  # noqa: E501
        :rtype: str
        """
        return self._aws_external_id

    @aws_external_id.setter
    def aws_external_id(self, aws_external_id):
        """Sets the aws_external_id of this ResponseOrganization.


        :param aws_external_id: The aws_external_id of this ResponseOrganization.  # noqa: E501
        :type aws_external_id: str
        """
        if self.local_vars_configuration.client_side_validation and aws_external_id is None:  # noqa: E501
            raise ValueError("Invalid value for `aws_external_id`, must not be `None`")  # noqa: E501

        self._aws_external_id = aws_external_id

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseOrganization.  # noqa: E501


        :return: The created_dt of this ResponseOrganization.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseOrganization.


        :param created_dt: The created_dt of this ResponseOrganization.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def credit_balance(self):
        """Gets the credit_balance of this ResponseOrganization.  # noqa: E501


        :return: The credit_balance of this ResponseOrganization.  # noqa: E501
        :rtype: float
        """
        return self._credit_balance

    @credit_balance.setter
    def credit_balance(self, credit_balance):
        """Sets the credit_balance of this ResponseOrganization.


        :param credit_balance: The credit_balance of this ResponseOrganization.  # noqa: E501
        :type credit_balance: float
        """
        if self.local_vars_configuration.client_side_validation and credit_balance is None:  # noqa: E501
            raise ValueError("Invalid value for `credit_balance`, must not be `None`")  # noqa: E501

        self._credit_balance = credit_balance

    @property
    def description(self):
        """Gets the description of this ResponseOrganization.  # noqa: E501


        :return: The description of this ResponseOrganization.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponseOrganization.


        :param description: The description of this ResponseOrganization.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def display_name(self):
        """Gets the display_name of this ResponseOrganization.  # noqa: E501


        :return: The display_name of this ResponseOrganization.  # noqa: E501
        :rtype: str
        """
        return self._display_name

    @display_name.setter
    def display_name(self, display_name):
        """Sets the display_name of this ResponseOrganization.


        :param display_name: The display_name of this ResponseOrganization.  # noqa: E501
        :type display_name: str
        """
        if self.local_vars_configuration.client_side_validation and display_name is None:  # noqa: E501
            raise ValueError("Invalid value for `display_name`, must not be `None`")  # noqa: E501

        self._display_name = display_name

    @property
    def id(self):
        """Gets the id of this ResponseOrganization.  # noqa: E501


        :return: The id of this ResponseOrganization.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseOrganization.


        :param id: The id of this ResponseOrganization.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this ResponseOrganization.  # noqa: E501


        :return: The name of this ResponseOrganization.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseOrganization.


        :param name: The name of this ResponseOrganization.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def owner_id(self):
        """Gets the owner_id of this ResponseOrganization.  # noqa: E501


        :return: The owner_id of this ResponseOrganization.  # noqa: E501
        :rtype: int
        """
        return self._owner_id

    @owner_id.setter
    def owner_id(self, owner_id):
        """Sets the owner_id of this ResponseOrganization.


        :param owner_id: The owner_id of this ResponseOrganization.  # noqa: E501
        :type owner_id: int
        """

        self._owner_id = owner_id

    @property
    def pricing_plan(self):
        """Gets the pricing_plan of this ResponseOrganization.  # noqa: E501


        :return: The pricing_plan of this ResponseOrganization.  # noqa: E501
        :rtype: ResponsePricingPlan
        """
        return self._pricing_plan

    @pricing_plan.setter
    def pricing_plan(self, pricing_plan):
        """Sets the pricing_plan of this ResponseOrganization.


        :param pricing_plan: The pricing_plan of this ResponseOrganization.  # noqa: E501
        :type pricing_plan: ResponsePricingPlan
        """
        if self.local_vars_configuration.client_side_validation and pricing_plan is None:  # noqa: E501
            raise ValueError("Invalid value for `pricing_plan`, must not be `None`")  # noqa: E501

        self._pricing_plan = pricing_plan

    @property
    def primary_owner_id(self):
        """Gets the primary_owner_id of this ResponseOrganization.  # noqa: E501


        :return: The primary_owner_id of this ResponseOrganization.  # noqa: E501
        :rtype: int
        """
        return self._primary_owner_id

    @primary_owner_id.setter
    def primary_owner_id(self, primary_owner_id):
        """Sets the primary_owner_id of this ResponseOrganization.


        :param primary_owner_id: The primary_owner_id of this ResponseOrganization.  # noqa: E501
        :type primary_owner_id: int
        """
        if self.local_vars_configuration.client_side_validation and primary_owner_id is None:  # noqa: E501
            raise ValueError("Invalid value for `primary_owner_id`, must not be `None`")  # noqa: E501

        self._primary_owner_id = primary_owner_id

    @property
    def show_tutorial(self):
        """Gets the show_tutorial of this ResponseOrganization.  # noqa: E501


        :return: The show_tutorial of this ResponseOrganization.  # noqa: E501
        :rtype: bool
        """
        return self._show_tutorial

    @show_tutorial.setter
    def show_tutorial(self, show_tutorial):
        """Sets the show_tutorial of this ResponseOrganization.


        :param show_tutorial: The show_tutorial of this ResponseOrganization.  # noqa: E501
        :type show_tutorial: bool
        """
        if self.local_vars_configuration.client_side_validation and show_tutorial is None:  # noqa: E501
            raise ValueError("Invalid value for `show_tutorial`, must not be `None`")  # noqa: E501

        self._show_tutorial = show_tutorial

    @property
    def stripe_customer_id(self):
        """Gets the stripe_customer_id of this ResponseOrganization.  # noqa: E501


        :return: The stripe_customer_id of this ResponseOrganization.  # noqa: E501
        :rtype: str
        """
        return self._stripe_customer_id

    @stripe_customer_id.setter
    def stripe_customer_id(self, stripe_customer_id):
        """Sets the stripe_customer_id of this ResponseOrganization.


        :param stripe_customer_id: The stripe_customer_id of this ResponseOrganization.  # noqa: E501
        :type stripe_customer_id: str
        """

        self._stripe_customer_id = stripe_customer_id

    @property
    def stripe_subscription_id(self):
        """Gets the stripe_subscription_id of this ResponseOrganization.  # noqa: E501


        :return: The stripe_subscription_id of this ResponseOrganization.  # noqa: E501
        :rtype: str
        """
        return self._stripe_subscription_id

    @stripe_subscription_id.setter
    def stripe_subscription_id(self, stripe_subscription_id):
        """Sets the stripe_subscription_id of this ResponseOrganization.


        :param stripe_subscription_id: The stripe_subscription_id of this ResponseOrganization.  # noqa: E501
        :type stripe_subscription_id: str
        """

        self._stripe_subscription_id = stripe_subscription_id

    @property
    def tutorials(self):
        """Gets the tutorials of this ResponseOrganization.  # noqa: E501


        :return: The tutorials of this ResponseOrganization.  # noqa: E501
        :rtype: list[ResponseTutorialResponse]
        """
        return self._tutorials

    @tutorials.setter
    def tutorials(self, tutorials):
        """Sets the tutorials of this ResponseOrganization.


        :param tutorials: The tutorials of this ResponseOrganization.  # noqa: E501
        :type tutorials: list[ResponseTutorialResponse]
        """
        if self.local_vars_configuration.client_side_validation and tutorials is None:  # noqa: E501
            raise ValueError("Invalid value for `tutorials`, must not be `None`")  # noqa: E501

        self._tutorials = tutorials

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseOrganization.  # noqa: E501


        :return: The updated_dt of this ResponseOrganization.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseOrganization.


        :param updated_dt: The updated_dt of this ResponseOrganization.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseOrganization):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseOrganization):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseMyUser(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'default_organization': 'ResponseOrganization',
        'display_name': 'str',
        'email': 'str',
        'git_ssh_key_name': 'str',
        'has_password': 'bool',
        'id': 'int',
        'is_email_verified': 'bool',
        'is_pending': 'bool',
        'is_superuser': 'bool',
        'last_login': 'datetime',
        'notification_config': 'dict[str, bool]',
        'organizations': 'list[ResponseOrganization]',
        'updated_dt': 'datetime',
        'username': 'str'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'default_organization': 'default_organization',
        'display_name': 'display_name',
        'email': 'email',
        'git_ssh_key_name': 'git_ssh_key_name',
        'has_password': 'has_password',
        'id': 'id',
        'is_email_verified': 'is_email_verified',
        'is_pending': 'is_pending',
        'is_superuser': 'is_superuser',
        'last_login': 'last_login',
        'notification_config': 'notification_config',
        'organizations': 'organizations',
        'updated_dt': 'updated_dt',
        'username': 'username'
    }

    def __init__(self, created_dt=None, default_organization=None, display_name=None, email=None, git_ssh_key_name=None, has_password=None, id=None, is_email_verified=None, is_pending=None, is_superuser=None, last_login=None, notification_config=None, organizations=None, updated_dt=None, username=None, local_vars_configuration=None):  # noqa: E501
        """ResponseMyUser - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._default_organization = None
        self._display_name = None
        self._email = None
        self._git_ssh_key_name = None
        self._has_password = None
        self._id = None
        self._is_email_verified = None
        self._is_pending = None
        self._is_superuser = None
        self._last_login = None
        self._notification_config = None
        self._organizations = None
        self._updated_dt = None
        self._username = None
        self.discriminator = None

        self.created_dt = created_dt
        self.default_organization = default_organization
        self.display_name = display_name
        self.email = email
        self.git_ssh_key_name = git_ssh_key_name
        self.has_password = has_password
        self.id = id
        self.is_email_verified = is_email_verified
        self.is_pending = is_pending
        self.is_superuser = is_superuser
        self.last_login = last_login
        self.notification_config = notification_config
        self.organizations = organizations
        self.updated_dt = updated_dt
        self.username = username

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseMyUser.  # noqa: E501


        :return: The created_dt of this ResponseMyUser.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseMyUser.


        :param created_dt: The created_dt of this ResponseMyUser.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def default_organization(self):
        """Gets the default_organization of this ResponseMyUser.  # noqa: E501


        :return: The default_organization of this ResponseMyUser.  # noqa: E501
        :rtype: ResponseOrganization
        """
        return self._default_organization

    @default_organization.setter
    def default_organization(self, default_organization):
        """Sets the default_organization of this ResponseMyUser.


        :param default_organization: The default_organization of this ResponseMyUser.  # noqa: E501
        :type default_organization: ResponseOrganization
        """
        if self.local_vars_configuration.client_side_validation and default_organization is None:  # noqa: E501
            raise ValueError("Invalid value for `default_organization`, must not be `None`")  # noqa: E501

        self._default_organization = default_organization

    @property
    def display_name(self):
        """Gets the display_name of this ResponseMyUser.  # noqa: E501


        :return: The display_name of this ResponseMyUser.  # noqa: E501
        :rtype: str
        """
        return self._display_name

    @display_name.setter
    def display_name(self, display_name):
        """Sets the display_name of this ResponseMyUser.


        :param display_name: The display_name of this ResponseMyUser.  # noqa: E501
        :type display_name: str
        """
        if self.local_vars_configuration.client_side_validation and display_name is None:  # noqa: E501
            raise ValueError("Invalid value for `display_name`, must not be `None`")  # noqa: E501

        self._display_name = display_name

    @property
    def email(self):
        """Gets the email of this ResponseMyUser.  # noqa: E501


        :return: The email of this ResponseMyUser.  # noqa: E501
        :rtype: str
        """
        return self._email

    @email.setter
    def email(self, email):
        """Sets the email of this ResponseMyUser.


        :param email: The email of this ResponseMyUser.  # noqa: E501
        :type email: str
        """
        if self.local_vars_configuration.client_side_validation and email is None:  # noqa: E501
            raise ValueError("Invalid value for `email`, must not be `None`")  # noqa: E501

        self._email = email

    @property
    def git_ssh_key_name(self):
        """Gets the git_ssh_key_name of this ResponseMyUser.  # noqa: E501


        :return: The git_ssh_key_name of this ResponseMyUser.  # noqa: E501
        :rtype: str
        """
        return self._git_ssh_key_name

    @git_ssh_key_name.setter
    def git_ssh_key_name(self, git_ssh_key_name):
        """Sets the git_ssh_key_name of this ResponseMyUser.


        :param git_ssh_key_name: The git_ssh_key_name of this ResponseMyUser.  # noqa: E501
        :type git_ssh_key_name: str
        """

        self._git_ssh_key_name = git_ssh_key_name

    @property
    def has_password(self):
        """Gets the has_password of this ResponseMyUser.  # noqa: E501


        :return: The has_password of this ResponseMyUser.  # noqa: E501
        :rtype: bool
        """
        return self._has_password

    @has_password.setter
    def has_password(self, has_password):
        """Sets the has_password of this ResponseMyUser.


        :param has_password: The has_password of this ResponseMyUser.  # noqa: E501
        :type has_password: bool
        """
        if self.local_vars_configuration.client_side_validation and has_password is None:  # noqa: E501
            raise ValueError("Invalid value for `has_password`, must not be `None`")  # noqa: E501

        self._has_password = has_password

    @property
    def id(self):
        """Gets the id of this ResponseMyUser.  # noqa: E501


        :return: The id of this ResponseMyUser.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseMyUser.


        :param id: The id of this ResponseMyUser.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def is_email_verified(self):
        """Gets the is_email_verified of this ResponseMyUser.  # noqa: E501


        :return: The is_email_verified of this ResponseMyUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_email_verified

    @is_email_verified.setter
    def is_email_verified(self, is_email_verified):
        """Sets the is_email_verified of this ResponseMyUser.


        :param is_email_verified: The is_email_verified of this ResponseMyUser.  # noqa: E501
        :type is_email_verified: bool
        """
        if self.local_vars_configuration.client_side_validation and is_email_verified is None:  # noqa: E501
            raise ValueError("Invalid value for `is_email_verified`, must not be `None`")  # noqa: E501

        self._is_email_verified = is_email_verified

    @property
    def is_pending(self):
        """Gets the is_pending of this ResponseMyUser.  # noqa: E501


        :return: The is_pending of this ResponseMyUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_pending

    @is_pending.setter
    def is_pending(self, is_pending):
        """Sets the is_pending of this ResponseMyUser.


        :param is_pending: The is_pending of this ResponseMyUser.  # noqa: E501
        :type is_pending: bool
        """
        if self.local_vars_configuration.client_side_validation and is_pending is None:  # noqa: E501
            raise ValueError("Invalid value for `is_pending`, must not be `None`")  # noqa: E501

        self._is_pending = is_pending

    @property
    def is_superuser(self):
        """Gets the is_superuser of this ResponseMyUser.  # noqa: E501


        :return: The is_superuser of this ResponseMyUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_superuser

    @is_superuser.setter
    def is_superuser(self, is_superuser):
        """Sets the is_superuser of this ResponseMyUser.


        :param is_superuser: The is_superuser of this ResponseMyUser.  # noqa: E501
        :type is_superuser: bool
        """
        if self.local_vars_configuration.client_side_validation and is_superuser is None:  # noqa: E501
            raise ValueError("Invalid value for `is_superuser`, must not be `None`")  # noqa: E501

        self._is_superuser = is_superuser

    @property
    def last_login(self):
        """Gets the last_login of this ResponseMyUser.  # noqa: E501


        :return: The last_login of this ResponseMyUser.  # noqa: E501
        :rtype: datetime
        """
        return self._last_login

    @last_login.setter
    def last_login(self, last_login):
        """Sets the last_login of this ResponseMyUser.


        :param last_login: The last_login of this ResponseMyUser.  # noqa: E501
        :type last_login: datetime
        """

        self._last_login = last_login

    @property
    def notification_config(self):
        """Gets the notification_config of this ResponseMyUser.  # noqa: E501


        :return: The notification_config of this ResponseMyUser.  # noqa: E501
        :rtype: dict[str, bool]
        """
        return self._notification_config

    @notification_config.setter
    def notification_config(self, notification_config):
        """Sets the notification_config of this ResponseMyUser.


        :param notification_config: The notification_config of this ResponseMyUser.  # noqa: E501
        :type notification_config: dict[str, bool]
        """
        if self.local_vars_configuration.client_side_validation and notification_config is None:  # noqa: E501
            raise ValueError("Invalid value for `notification_config`, must not be `None`")  # noqa: E501

        self._notification_config = notification_config

    @property
    def organizations(self):
        """Gets the organizations of this ResponseMyUser.  # noqa: E501


        :return: The organizations of this ResponseMyUser.  # noqa: E501
        :rtype: list[ResponseOrganization]
        """
        return self._organizations

    @organizations.setter
    def organizations(self, organizations):
        """Sets the organizations of this ResponseMyUser.


        :param organizations: The organizations of this ResponseMyUser.  # noqa: E501
        :type organizations: list[ResponseOrganization]
        """
        if self.local_vars_configuration.client_side_validation and organizations is None:  # noqa: E501
            raise ValueError("Invalid value for `organizations`, must not be `None`")  # noqa: E501

        self._organizations = organizations

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseMyUser.  # noqa: E501


        :return: The updated_dt of this ResponseMyUser.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseMyUser.


        :param updated_dt: The updated_dt of this ResponseMyUser.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    @property
    def username(self):
        """Gets the username of this ResponseMyUser.  # noqa: E501


        :return: The username of this ResponseMyUser.  # noqa: E501
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this ResponseMyUser.


        :param username: The username of this ResponseMyUser.  # noqa: E501
        :type username: str
        """
        if self.local_vars_configuration.client_side_validation and username is None:  # noqa: E501
            raise ValueError("Invalid value for `username`, must not be `None`")  # noqa: E501

        self._username = username

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseMyUser):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseMyUser):
            return True

        return self.to_dict() != other.to_dict()
